import argparse
import json
import os

import requests
from PIL import Image

__all__ = ["load_workflow_from_file", "queue_prompt"]

prompt_text = ""


def extract_metadata_from_png(png_file_path):
    img = Image.open(png_file_path)
    metadata = img.info
    img.close()
    return metadata


def load_workflow_from_file(filename) -> str:
    _, extension = os.path.splitext(filename)
    if extension.endswith("json"):
        with open(filename, "r") as f:
            c = f.read()
            return c
    elif extension.endswith("png"):
        metadata = extract_metadata_from_png(filename)
        return metadata["api"]
    else:
        raise NotImplementedError("Only json or png workflow file supported yet")


def parse_args():
    parser = argparse.ArgumentParser(description="Launch ComfyUI workflow by API")
    parser.add_argument(
        "-w",
        "--workflow",
        type=str,
        help="The workflow filename which can be either json or png.",
        required=True,
    )
    parser.add_argument(
        "--host",
        type=str,
        default="127.0.0.1",
        help="ComfyUI server host, default is 127.0.0.1",
    )
    parser.add_argument(
        "-p",
        "--port",
        type=int,
        default=8188,
        help="CofyUI server port, default is 8188",
    )

    args = parser.parse_args()
    return args


def queue_prompt(prompt, host, port):
    url = f"http://{host}:{port}/prompt"
    payload = {"prompt": prompt}
    headers = {"Content-Type": "application/json"}
    response = requests.post(url, data=json.dumps(payload), headers=headers)

    if response.status_code == 200:
        print("Prompt queued successfully.")
        result = json.loads(response.text)
        print(result)
        return result
    else:
        print(
            f"Failed to queue prompt, error code: {response.status_code}, {response.text}"
        )
    return None


if __name__ == "__main__":
    args = parse_args()
    r = load_workflow_from_file(args.workflow)
    prompt = json.loads(r)
    r = queue_prompt(prompt, args.host, args.port)



import uuid
from typing import Optional, Tuple

import folder_paths
import torch
from nodes import CheckpointLoaderSimple, ControlNetLoader
from onediff.utils.import_utils import (  # type: ignore
    is_nexfort_available,
    is_oneflow_available,
)

from ._config import is_disable_oneflow_backend
from .modules import BoosterExecutor, BoosterScheduler, BoosterSettings

if is_oneflow_available() and not is_disable_oneflow_backend():
    from .modules.oneflow import BasicOneFlowBoosterExecutor

    BasicBoosterExecutor = BasicOneFlowBoosterExecutor
    print("\033[1;31mUsing OneFlow backend\033[0m (Default)")
elif is_nexfort_available():
    from .modules.nexfort.booster_basic import BasicNexFortBoosterExecutor

    BasicBoosterExecutor = BasicNexFortBoosterExecutor
    print("\033[1;32mUsing Nexfort backend\033[0m (Default)")
else:
    raise RuntimeError(
        "Neither OneFlow nor Nexfort is available. Please ensure at least one of them is installed."
    )

__all__ = [
    "ModelSpeedup",
    "VaeSpeedup",
    "ControlnetSpeedup",
    "OneDiffApplyModelBooster",
    "OneDiffControlNetLoader",
    "OneDiffCheckpointLoaderSimple",
]


class SpeedupMixin:
    """A mix-in class to provide speedup functionality."""

    FUNCTION = "speedup"
    CATEGORY = "OneDiff"

    @torch.inference_mode()
    def speedup(
        self,
        model,
        ckpt_name: Optional[str] = None,
        inplace: bool = False,
        custom_booster: Optional[BoosterScheduler] = None,
        booster_settings: Optional[BoosterSettings] = None,
        *args,
        **kwargs
    ) -> Tuple:
        """
        Speed up the model inference.

        Args:
            model: The input model to be sped up.
            inplace (bool, optional): Whether to perform the operation inplace. Defaults to False.
            custom_booster (BoosterScheduler, optional): Custom booster scheduler to use. Defaults to None.
            *args: Additional positional arguments to be passed to the underlying functions.
            **kwargs: Additional keyword arguments to be passed to the underlying functions.

        Returns:
            Tuple: Tuple containing the optimized model.
        """
        if booster_settings is None and not hasattr(self, "booster_settings"):
            self.booster_settings = BoosterSettings(tmp_cache_key=str(uuid.uuid4()))

        if custom_booster:
            booster = custom_booster
            booster.inplace = inplace
        else:
            booster = BoosterScheduler(BasicBoosterExecutor(), inplace=inplace)
        booster.settings = (
            self.booster_settings if booster_settings is None else booster_settings
        )
        return (booster(model, ckpt_name, *args, **kwargs),)


class ModelSpeedup(SpeedupMixin):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "model": ("MODEL",),
            },
            "optional": {
                "custom_booster": ("CUSTOM_BOOSTER",),
                "inplace": (
                    "BOOLEAN",
                    {"default": True, "label_on": "yes", "label_off": "no"},
                ),
            },
        }

    RETURN_TYPES = ("MODEL",)


class VaeSpeedup(SpeedupMixin):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {"vae": ("VAE",)},
            "optional": {
                "custom_booster": ("CUSTOM_BOOSTER",),
                "inplace": (
                    "BOOLEAN",
                    {"default": True, "label_on": "yes", "label_off": "no"},
                ),
            },
        }

    RETURN_TYPES = ("VAE",)

    def speedup(self, vae, inplace=False, custom_booster: BoosterScheduler = None):
        return super().speedup(vae, inplace, custom_booster)


class ControlnetSpeedup(SpeedupMixin):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "control_net": ("CONTROL_NET",),
            },
            "optional": {
                "inplace": (
                    "BOOLEAN",
                    {"default": True, "label_on": "yes", "label_off": "no"},
                ),
                "custom_booster": ("CUSTOM_BOOSTER",),
            },
        }

    RETURN_TYPES = ("CONTROL_NET",)
    FUNCTION = "speedup"
    CATEGORY = "OneDiff"

    def speedup(
        self,
        control_net=None,
        inplace=True,
        custom_booster: BoosterScheduler = None,
        **kwargs
    ):
        return super().speedup(control_net, inplace, custom_booster)


class OneDiffApplyModelBooster:
    """Main class responsible for optimizing models."""

    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {},
            "optional": {
                "quantization_booster": ("QuantizationBooster",),
                "deepcache_booster": ("DeepCacheBooster",),
                "torchcompile_booster": ("TorchCompileBooster",),
            },
        }

    CATEGORY = "OneDiff/Booster"
    RETURN_TYPES = ("CUSTOM_BOOSTER",)
    FUNCTION = "speedup_module"

    @torch.no_grad()
    def speedup_module(
        self,
        quantization_booster: BoosterExecutor = None,
        deepcache_booster=None,
        torchcompile_booster=None,
    ):
        """Apply the optimization technique to the model."""
        booster_executors = []
        if deepcache_booster:
            booster_executors.append(deepcache_booster)
        if quantization_booster:
            booster_executors.append(quantization_booster)
        if torchcompile_booster:
            booster_executors.append(torchcompile_booster)

        assert len(booster_executors) > 0
        return (BoosterScheduler(booster_executors),)


class OneDiffControlNetLoader(ControlNetLoader):
    @classmethod
    def INPUT_TYPES(s):
        ret = super().INPUT_TYPES()
        ret.update(
            {
                "optional": {
                    "custom_booster": ("CUSTOM_BOOSTER",),
                }
            }
        )
        return ret

    CATEGORY = "OneDiff/Loaders"
    FUNCTION = "onediff_load_controlnet"

    @torch.inference_mode()
    def onediff_load_controlnet(self, control_net_name, custom_booster=None):
        controlnet = super().load_controlnet(control_net_name)[0]
        if custom_booster is None:
            custom_booster = BoosterScheduler(BasicBoosterExecutor())
        controlnet = custom_booster(controlnet, ckpt_name=control_net_name)

        return (controlnet,)


class OneDiffCheckpointLoaderSimple(CheckpointLoaderSimple, SpeedupMixin):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "ckpt_name": (folder_paths.get_filename_list("checkpoints"),),
                "vae_speedup": (["disable", "enable"],),
            },
            "optional": {
                "custom_booster": ("CUSTOM_BOOSTER",),
            },
        }

    CATEGORY = "OneDiff/Loaders"
    FUNCTION = "onediff_load_checkpoint"

    def __init__(self) -> None:
        super().__init__()
        self.unet_booster_settings = BoosterSettings(tmp_cache_key=str(uuid.uuid4()))
        self.vae_booster_settings = BoosterSettings(tmp_cache_key=str(uuid.uuid4()))

    @torch.inference_mode()
    def onediff_load_checkpoint(
        self,
        ckpt_name,
        vae_speedup="disable",
        custom_booster: BoosterScheduler = None,
    ):
        modelpatcher, clip, vae = self.load_checkpoint(ckpt_name)
        modelpatcher = self.speedup(
            modelpatcher,
            ckpt_name=ckpt_name,
            inplace=True,
            custom_booster=custom_booster,
            booster_settings=self.unet_booster_settings,
        )[0]

        if vae_speedup == "enable":
            vae = self.speedup(
                vae,
                inplace=True,
                custom_booster=custom_booster,
                booster_settings=self.vae_booster_settings,
            )[0]

        # Set weight inplace update
        modelpatcher.weight_inplace_update = True
        return (
            modelpatcher,
            clip,
            vae,
        )



from pathlib import Path

import folder_paths
import torch
from comfy import model_management
from onediff.infer_compiler import oneflow_compile, OneflowCompileOptions

from ..config import _USE_UNET_INT8, ONEDIFF_QUANTIZED_OPTIMIZED_MODELS
from .graph_path import generate_graph_path
from .loader_sample_tools import compoile_unet, quantize_unet
from .model_patcher import OneFlowSpeedUpModelPatcher


def onediff_load_quant_checkpoint_advanced(
    ckpt_name, model_path, need_compile, vae_speedup, modelpatcher, vae
):
    """
    Loads a quantized and potentially optimized model checkpoint, applies quantization,
    optionally compiles the model, and applies VAE speedup if enabled.
    """
    ckpt_name = f"{ckpt_name}_quant_{model_path}"
    model_path = (
        Path(folder_paths.models_dir) / ONEDIFF_QUANTIZED_OPTIMIZED_MODELS / model_path
    )
    graph_file = generate_graph_path(ckpt_name, modelpatcher.model)

    calibrate_info = torch.load(model_path)

    load_device = model_management.get_torch_device()
    diffusion_model = modelpatcher.model.diffusion_model.to(load_device)
    quant_unet = quantize_unet(
        diffusion_model=diffusion_model,
        inplace=True,
        calibrate_info=calibrate_info,
    )
    modelpatcher.model.diffusion_model = quant_unet

    if need_compile:
        # compiled_unet = compoile_unet(
        #     modelpatcher.model.diffusion_model, graph_file
        # )
        # modelpatcher.model.diffusion_model = compiled_unet
        offload_device = model_management.unet_offload_device()
        modelpatcher = OneFlowSpeedUpModelPatcher(
            modelpatcher.model,
            load_device=model_management.get_torch_device(),
            offload_device=offload_device,
            use_graph=True,
            graph_path=graph_file,
            graph_device=model_management.get_torch_device(),
        )

    if vae_speedup == "enable":
        compile_options = OneflowCompileOptions()
        compile_options.graph_file = generate_graph_path(
            ckpt_name, vae.first_stage_model
        )
        compile_options.graph_file_device = model_management.get_torch_device()
        vae.first_stage_model = oneflow_compile(
            vae.first_stage_model, options=compile_options
        )

    # set inplace update
    modelpatcher.weight_inplace_update = True
    return modelpatcher, vae



import os
from pathlib import Path

import comfy
import folder_paths
import torch
from comfy import model_management
from comfy.cli_args import args
from onediff.infer_compiler.backends.oneflow.utils.version_util import (
    is_community_version,
)

from onediff.utils.import_utils import is_onediff_quant_available

from ..modules import BoosterScheduler
from ..modules.oneflow import (
    BasicOneFlowBoosterExecutor,
    DeepcacheBoosterExecutor,
    PatchBoosterExecutor,
)
from ..modules.oneflow.config import ONEDIFF_QUANTIZED_OPTIMIZED_MODELS
from ..modules.oneflow.utils import load_graph, OUTPUT_FOLDER, save_graph

if is_onediff_quant_available() and not is_community_version():
    from ..modules.oneflow.booster_quantization import (
        OnelineQuantizationBoosterExecutor,
    )  # type: ignore


import comfy_extras.nodes_video_model
from nodes import CheckpointLoaderSimple

# https://github.com/comfyanonymous/ComfyUI/commit/bb4940d837f0cfd338ff64776b084303be066c67#diff-fab3fbd81daf87571b12fb3e4d80fc7d6bbbcf0f3dafed1dbc55d81998d82539L54
if hasattr(args, "dont_upcast_attention") and not args.dont_upcast_attention:
    os.environ["ONEFLOW_ATTENTION_ALLOW_HALF_PRECISION_SCORE_ACCUMULATION_MAX_M"] = "0"


class OneFlowDeepcacheBooster:
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "cache_interval": (
                    "INT",
                    {
                        "default": 3,
                        "min": 1,
                        "max": 1000,
                        "step": 1,
                        "display": "number",
                    },
                ),
                "cache_layer_id": (
                    "INT",
                    {"default": 0, "min": 0, "max": 12, "step": 1, "display": "number"},
                ),
                "cache_block_id": (
                    "INT",
                    {"default": 1, "min": 0, "max": 12, "step": 1, "display": "number"},
                ),
                "start_step": (
                    "INT",
                    {
                        "default": 0,
                        "min": 0,
                        "max": 1000,
                        "step": 1,
                        "display": "number",
                    },
                ),
                "end_step": (
                    "INT",
                    {"default": 1000, "min": 0, "max": 1000, "step": 0.1},
                ),
            },
        }

    CATEGORY = "OneDiff/Booster"
    RETURN_TYPES = ("DeepCacheBooster",)
    FUNCTION = "apply"

    @torch.no_grad()
    def apply(
        self,
        cache_interval=3,
        cache_layer_id=0,
        cache_block_id=1,
        start_step=0,
        end_step=1000,
    ):
        return (
            DeepcacheBoosterExecutor(
                cache_interval=cache_interval,
                cache_layer_id=cache_layer_id,
                cache_block_id=cache_block_id,
                start_step=start_step,
                end_step=end_step,
            ),
        )


class ModuleDeepCacheSpeedup:
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "model": ("MODEL",),
                "static_mode": (["enable", "disable"],),
                "cache_interval": (
                    "INT",
                    {
                        "default": 3,
                        "min": 1,
                        "max": 1000,
                        "step": 1,
                        "display": "number",
                    },
                ),
                "cache_layer_id": (
                    "INT",
                    {"default": 0, "min": 0, "max": 12, "step": 1, "display": "number"},
                ),
                "cache_block_id": (
                    "INT",
                    {"default": 1, "min": 0, "max": 12, "step": 1, "display": "number"},
                ),
                "start_step": (
                    "INT",
                    {
                        "default": 0,
                        "min": 0,
                        "max": 1000,
                        "step": 1,
                        "display": "number",
                    },
                ),
                "end_step": (
                    "INT",
                    {
                        "default": 1000,
                        "min": 0,
                        "max": 1000,
                        "step": 0.1,
                    },
                ),
            },
        }

    RETURN_TYPES = ("MODEL",)
    FUNCTION = "deep_cache_convert"
    CATEGORY = "OneDiff"

    @torch.no_grad()
    def deep_cache_convert(
        self,
        model,
        static_mode,
        cache_interval,
        cache_layer_id,
        cache_block_id,
        start_step,
        end_step,
    ):
        print(
            f"Warning: {type(self).__name__} will be deleted. Please use it with caution."
        )
        booster = BoosterScheduler(
            DeepcacheBoosterExecutor(
                cache_interval=cache_interval,
                cache_layer_id=cache_layer_id,
                cache_block_id=cache_block_id,
                start_step=start_step,
                end_step=end_step,
            )
        )

        return (booster(model),)


class OneDiffOnlineQuantizationBooster:
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "quantized_conv_percentage": (
                    "INT",
                    {
                        "default": 100,
                        "min": 0,  # Minimum value
                        "max": 100,  # Maximum value
                        "step": 1,  # Slider's step
                        "display": "slider",  # Cosmetic only: display as "number" or "slider"
                    },
                ),
                "quantized_linear_percentage": (
                    "INT",
                    {
                        "default": 100,
                        "min": 0,  # Minimum value
                        "max": 100,  # Maximum value
                        "step": 1,  # Slider's step
                        "display": "slider",  # Cosmetic only: display as "number" or "slider"
                    },
                ),
                "conv_compute_density_threshold": (
                    "INT",
                    {
                        "default": 100,
                        "min": 0,  # Minimum value
                        "max": 2000,  # Maximum value
                        "step": 10,  # Slider's step
                        "display": "number",  # Cosmetic only: display as "number" or "slider"
                    },
                ),
                "linear_compute_density_threshold": (
                    "INT",
                    {
                        "default": 300,
                        "min": 0,  # Minimum value
                        "max": 2000,  # Maximum value
                        "step": 10,  # Slider's step
                        "display": "number",  # Cosmetic only: display as "number" or "slider"
                    },
                ),
            },
        }

    CATEGORY = "OneDiff/Booster"
    RETURN_TYPES = ("QuantizationBooster",)
    FUNCTION = "apply"

    @torch.no_grad()
    def apply(
        self,
        quantized_conv_percentage=0,
        quantized_linear_percentage=0,
        conv_compute_density_threshold=0,
        linear_compute_density_threshold=0,
    ):
        if not is_onediff_quant_available() or is_community_version():
            raise RuntimeError(
                f"OneDiff Quant or community version are not available. "
                f"Please refer to the documentation for reinstalling OneDiff Enterprise: "
                f"https://github.com/siliconflow/onediff/blob/main/README_ENTERPRISE.md#install-onediff-enterprise\n"
                f"is_community_version={is_community_version()}\n"
                f"is_onediff_quant_available={is_onediff_quant_available()}"
            )

        return (
            OnelineQuantizationBoosterExecutor(
                conv_percentage=quantized_conv_percentage,
                linear_percentage=quantized_linear_percentage,
                conv_compute_density_threshold=conv_compute_density_threshold,
                linear_compute_density_threshold=linear_compute_density_threshold,
            ),
        )


class OneDiffDeepCacheCheckpointLoaderSimple(CheckpointLoaderSimple):
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "ckpt_name": (folder_paths.get_filename_list("checkpoints"),),
                "vae_speedup": (["disable", "enable"],),
                "static_mode": (["enable", "disable"],),
                "cache_interval": (
                    "INT",
                    {
                        "default": 3,
                        "min": 1,
                        "max": 1000,
                        "step": 1,
                        "display": "number",
                    },
                ),
                "cache_layer_id": (
                    "INT",
                    {"default": 0, "min": 0, "max": 12, "step": 1, "display": "number"},
                ),
                "cache_block_id": (
                    "INT",
                    {"default": 1, "min": 0, "max": 12, "step": 1, "display": "number"},
                ),
                "start_step": (
                    "INT",
                    {
                        "default": 0,
                        "min": 0,
                        "max": 1000,
                        "step": 1,
                        "display": "number",
                    },
                ),
                "end_step": (
                    "INT",
                    {
                        "default": 1000,
                        "min": 0,
                        "max": 1000,
                        "step": 0.1,
                    },
                ),
            }
        }

    CATEGORY = "OneDiff/Loaders"
    FUNCTION = "onediff_load_checkpoint"

    @torch.no_grad()
    def onediff_load_checkpoint(
        self,
        ckpt_name,
        vae_speedup,
        static_mode="enable",
        cache_interval=3,
        cache_layer_id=0,
        cache_block_id=1,
        start_step=0,
        end_step=1000,
    ):
        print(
            f"Warning: {type(self).__name__} will be deleted. Please use it with caution."
        )
        # CheckpointLoaderSimple.load_checkpoint
        modelpatcher, clip, vae = self.load_checkpoint(ckpt_name)
        booster = BoosterScheduler(
            DeepcacheBoosterExecutor(
                cache_interval=cache_interval,
                cache_layer_id=cache_layer_id,
                cache_block_id=cache_block_id,
                start_step=start_step,
                end_step=end_step,
            )
        )

        modelpatcher = booster(modelpatcher, ckpt_name=ckpt_name)
        if vae_speedup == "enable":
            vae = BoosterScheduler(BasicOneFlowBoosterExecutor())(
                vae, ckpt_name=ckpt_name
            )

        # set inplace update
        modelpatcher.weight_inplace_update = True
        return modelpatcher, clip, vae


class BatchSizePatcher:
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "model": ("MODEL",),
                "latent_image": ("LATENT",),
            },
        }

    RETURN_TYPES = ("MODEL",)
    CATEGORY = "OneDiff/Tools"
    FUNCTION = "set_cache_filename"

    @torch.no_grad()
    def set_cache_filename(self, model, latent_image):
        booster = BoosterScheduler(PatchBoosterExecutor())
        model = booster(model=model, latent_image=latent_image)
        return (model,)


class SVDSpeedup:
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "model": ("MODEL",),
                "inplace": ([False, True],),
                "cache_name": (
                    "STRING",
                    {
                        "multiline": False,  # True if you want the field to look like the one on the ClipTextEncode node
                        "default": "svd",
                    },
                ),
            },
            "optional": {
                "custom_booster": ("CUSTOM_BOOSTER",),
            },
        }

    RETURN_TYPES = ("MODEL",)
    FUNCTION = "speedup"
    CATEGORY = "OneDiff"

    @torch.no_grad()
    def speedup(
        self,
        model,
        inplace=False,
        cache_name="svd",
        custom_booster: BoosterScheduler = None,
    ):
        print(
            f"Warning: {type(self).__name__} will be deleted. Please use it with caution."
        )
        if custom_booster:
            booster = custom_booster
            booster.inplace = inplace
        else:
            booster = BoosterScheduler(BasicOneFlowBoosterExecutor(), inplace=inplace)

        optimized_model = booster.compile(model, ckpt_name=cache_name)
        return (optimized_model,)


########################## For downward compatibility, it is retained ###################
class VaeGraphLoader:
    @classmethod
    def INPUT_TYPES(s):
        vae_folder = os.path.join(OUTPUT_FOLDER, "vae")
        graph_files = [
            f
            for f in os.listdir(vae_folder)
            if os.path.isfile(os.path.join(vae_folder, f)) and f.endswith(".graph")
        ]
        return {
            "required": {
                "vae": ("VAE",),
                "graph": (sorted(graph_files),),
            },
        }

    RETURN_TYPES = ("VAE",)
    FUNCTION = "load_graph"
    CATEGORY = "OneDiff"

    def load_graph(self, vae, graph):
        print(
            f"Warning: {type(self).__name__} will be deleted. Please use it with caution."
        )
        vae_model = vae.first_stage_model
        device = model_management.vae_offload_device()
        load_graph(vae_model, graph, device, subfolder="vae")
        return (vae,)


class VaeGraphSaver:
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "images": ("IMAGE",),
                "vae": ("VAE",),
                "filename_prefix": ("STRING", {"default": "vae"}),
            },
        }

    RETURN_TYPES = ()
    FUNCTION = "save_graph"
    CATEGORY = "OneDiff"
    OUTPUT_NODE = True

    def save_graph(self, images, vae, filename_prefix):
        print(
            f"Warning: {type(self).__name__} will be deleted. Please use it with caution."
        )
        vae_model = vae.first_stage_model
        vae_device = model_management.vae_offload_device()
        save_graph(vae_model, filename_prefix, vae_device, subfolder="vae")

        return {}


class ModelGraphLoader:
    @classmethod
    def INPUT_TYPES(s):
        unet_folder = os.path.join(OUTPUT_FOLDER, "unet")
        graph_files = [
            f
            for f in os.listdir(unet_folder)
            if os.path.isfile(os.path.join(unet_folder, f)) and f.endswith(".graph")
        ]
        return {
            "required": {
                "model": ("MODEL",),
                "graph": (sorted(graph_files),),
            },
        }

    RETURN_TYPES = ("MODEL",)
    FUNCTION = "load_graph"
    CATEGORY = "OneDiff"

    def load_graph(self, model, graph):
        print(
            f"Warning: {type(self).__name__} will be deleted. Please use it with caution."
        )

        diffusion_model = model.model.diffusion_model

        load_graph(diffusion_model, graph, "cuda", subfolder="unet")
        return (model,)


class ModelGraphSaver:
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "samples": ("LATENT",),
                "model": ("MODEL",),
                "filename_prefix": ("STRING", {"default": "unet"}),
            },
        }

    RETURN_TYPES = ()
    FUNCTION = "save_graph"
    CATEGORY = "OneDiff"
    OUTPUT_NODE = True

    def save_graph(self, samples, model, filename_prefix):
        print(
            f"Warning: {type(self).__name__} will be deleted. Please use it with caution."
        )
        diffusion_model = model.model.diffusion_model
        save_graph(diffusion_model, filename_prefix, "cuda", subfolder="unet")
        return {}


NODE_CLASS_MAPPINGS = {
    "ModelGraphLoader": ModelGraphLoader,
    "ModelGraphSaver": ModelGraphSaver,
    "VaeGraphSaver": VaeGraphSaver,
    "VaeGraphLoader": VaeGraphLoader,
    "ModuleDeepCacheSpeedup": ModuleDeepCacheSpeedup,
    "OneDiffDeepCacheCheckpointLoaderSimple": OneDiffDeepCacheCheckpointLoaderSimple,
    "BatchSizePatcher": BatchSizePatcher,
    "OneDiffOnlineQuantizationBooster": OneDiffOnlineQuantizationBooster,
    "SVDSpeedup": SVDSpeedup,
    "OneFlowDeepcacheBooster": OneFlowDeepcacheBooster,
}

NODE_DISPLAY_NAME_MAPPINGS = {
    "ModelGraphLoader": "Model Graph Loader",
    "ModelGraphSaver": "Model Graph Saver",
    "VaeGraphLoader": "VAE Graph Loader",
    "VaeGraphSaver": "VAE Graph Saver",
    "SVDSpeedup": "SVD Speedup",
    "ModuleDeepCacheSpeedup": "Model DeepCache Speedup",
    "OneDiffControlNetLoader": "Load ControlNet Model - OneDiff",
    "OneDiffDeepCacheCheckpointLoaderSimple": "Load Checkpoint - OneDiff DeepCache",
    "BatchSizePatcher": "Batch Size Patcher",
    "OneDiffOnlineQuantizationBooster": "Online OneFlow Quantizer - OneDiff",
    "OneFlowDeepcacheBooster": "OneFlow Deepcache Booster - OneDiff",
}


if is_onediff_quant_available() and not is_community_version():

    class UNETLoaderInt8:
        @classmethod
        def INPUT_TYPES(cls):
            paths = []
            for search_path in folder_paths.get_folder_paths("unet_int8"):
                if os.path.exists(search_path):
                    for root, subdir, files in os.walk(search_path, followlinks=True):
                        if "calibrate_info.txt" in files:
                            paths.append(os.path.relpath(root, start=search_path))

            return {
                "required": {
                    "model_path": (paths,),
                }
            }

        RETURN_TYPES = ("MODEL",)
        FUNCTION = "load_unet_int8"

        CATEGORY = "OneDiff"

        def load_unet_int8(self, model_path):
            print(
                f"Warning: {type(self).__name__} will be deleted. Please use it with caution."
            )
            from ..modules.oneflow.utils.onediff_quant_utils import (
                replace_module_with_quantizable_module,
            )

            for search_path in folder_paths.get_folder_paths("unet_int8"):
                if os.path.exists(search_path):
                    path = os.path.join(search_path, model_path)
                    if os.path.exists(path):
                        model_path = path
                        break

            unet_sd_path = os.path.join(model_path, "unet_int8.safetensors")
            calibrate_info_path = os.path.join(model_path, "calibrate_info.txt")

            model = comfy.sd.load_unet(unet_sd_path)
            replace_module_with_quantizable_module(
                model.model.diffusion_model, calibrate_info_path
            )
            return (model,)

    class Quant8Model:
        @classmethod
        def INPUT_TYPES(s):
            return {
                "required": {
                    "model": ("MODEL",),
                    "output_dir": ("STRING", {"default": "int8"}),
                    "conv": (["enable", "disable"],),
                    "linear": (["enable", "disable"],),
                },
            }

        RETURN_TYPES = ()
        FUNCTION = "quantize_model"
        CATEGORY = "OneDiff"
        OUTPUT_NODE = True

        def quantize_model(self, model, output_dir, conv, linear):
            print(
                f"Warning: {type(self).__name__} will be deleted. Please use it with caution."
            )
            from ..modules.oneflow.utils import quantize_and_save_model

            diffusion_model = model.model.diffusion_model
            output_dir = os.path.join(folder_paths.models_dir, "unet_int8", output_dir)
            is_quantize_conv = conv == "enable"
            is_quantize_linear = linear == "enable"
            quantize_and_save_model(
                diffusion_model,
                output_dir,
                quantize_conv=is_quantize_conv,
                quantize_linear=is_quantize_linear,
                verbose=False,
            )
            return {}

    class OneDiffQuantCheckpointLoaderSimple(CheckpointLoaderSimple):
        @classmethod
        def INPUT_TYPES(s):
            return {
                "required": {
                    "ckpt_name": (folder_paths.get_filename_list("checkpoints"),),
                    "vae_speedup": (["disable", "enable"],),
                }
            }

        CATEGORY = "OneDiff/Loaders"
        FUNCTION = "onediff_load_checkpoint"

        def onediff_load_checkpoint(self, ckpt_name, vae_speedup):
            modelpatcher, clip, vae = self.load_checkpoint(ckpt_name)
            print(
                f"Warning: {type(self).__name__} will be deleted. Please use it with caution."
            )
            booster = BoosterScheduler(
                OnelineQuantizationBoosterExecutor(
                    conv_percentage=100,
                    linear_percentage=100,
                    conv_compute_density_threshold=600,
                    linear_compute_density_threshold=900,
                )
            )
            modelpatcher = booster.compile(modelpatcher, ckpt_name=ckpt_name)
            if vae_speedup == "enable":
                vae = BoosterScheduler(BasicOneFlowBoosterExecutor())(
                    vae, ckpt_name=ckpt_name
                )

            # set inplace update
            modelpatcher.weight_inplace_update = True
            return modelpatcher, clip, vae

    class OneDiffQuantCheckpointLoaderSimpleAdvanced(CheckpointLoaderSimple):
        @classmethod
        def INPUT_TYPES(s):
            paths = []
            for search_path in folder_paths.get_folder_paths(
                ONEDIFF_QUANTIZED_OPTIMIZED_MODELS
            ):
                if os.path.exists(search_path):
                    search_path = Path(search_path)
                    paths.extend(
                        [
                            os.path.relpath(p, start=search_path)
                            for p in search_path.glob("*.pt")
                        ]
                    )

            return {
                "required": {
                    "ckpt_name": (folder_paths.get_filename_list("checkpoints"),),
                    "model_path": (paths,),
                    "compile": (["enable", "disable"],),
                    "vae_speedup": (["disable", "enable"],),
                }
            }

        CATEGORY = "OneDiff/Loaders"
        FUNCTION = "onediff_load_checkpoint"

        def onediff_load_checkpoint(
            self,
            ckpt_name,
            model_path,
            compile,
            vae_speedup,
        ):
            need_compile = compile == "enable"
            print(
                f"Warning: {type(self).__name__} will be deleted. Please use it with caution."
            )

            modelpatcher, clip, vae = self.load_checkpoint(ckpt_name)
            # TODO fix by op.compile
            from ..modules.oneflow.utils.onediff_load_utils import (
                onediff_load_quant_checkpoint_advanced,
            )

            modelpatcher, vae = onediff_load_quant_checkpoint_advanced(
                ckpt_name, model_path, need_compile, vae_speedup, modelpatcher, vae
            )

            return modelpatcher, clip, vae

    class ImageOnlyOneDiffQuantCheckpointLoaderAdvanced(
        comfy_extras.nodes_video_model.ImageOnlyCheckpointLoader
    ):
        @classmethod
        def INPUT_TYPES(s):
            paths = []
            for search_path in folder_paths.get_folder_paths(
                ONEDIFF_QUANTIZED_OPTIMIZED_MODELS
            ):
                if os.path.exists(search_path):
                    search_path = Path(search_path)
                    paths.extend(
                        [
                            os.path.relpath(p, start=search_path)
                            for p in search_path.glob("*.pt")
                        ]
                    )

            return {
                "required": {
                    "ckpt_name": (folder_paths.get_filename_list("checkpoints"),),
                    "model_path": (paths,),
                    "compile": (["enable", "disable"],),
                    "vae_speedup": (["disable", "enable"],),
                }
            }

        CATEGORY = "OneDiff/Loaders"
        FUNCTION = "onediff_load_checkpoint"

        def onediff_load_checkpoint(
            self,
            ckpt_name,
            model_path,
            compile,
            vae_speedup,
            output_vae=True,
            output_clip=True,
        ):
            print(
                f"Warning: {type(self).__name__} will be deleted. Please use it with caution."
            )
            modelpatcher, clip, vae = self.load_checkpoint(
                ckpt_name, output_vae, output_clip
            )
            booster = BoosterScheduler(BasicOneFlowBoosterExecutor())
            modelpatcher = booster(modelpatcher, ckpt_name=ckpt_name)
            if vae_speedup:
                vae = BoosterScheduler(BasicOneFlowBoosterExecutor())(
                    vae, ckpt_name=ckpt_name
                )
            return modelpatcher, clip, vae

    NODE_CLASS_MAPPINGS.update(
        {
            "UNETLoaderInt8": UNETLoaderInt8,
            "Quant8Model": Quant8Model,
            "OneDiffQuantCheckpointLoaderSimple": OneDiffQuantCheckpointLoaderSimple,
            "OneDiffQuantCheckpointLoaderSimpleAdvanced": OneDiffQuantCheckpointLoaderSimpleAdvanced,
            "ImageOnlyOneDiffQuantCheckpointLoaderAdvanced": ImageOnlyOneDiffQuantCheckpointLoaderAdvanced,
        }
    )

    NODE_DISPLAY_NAME_MAPPINGS.update(
        {
            "UNETLoaderInt8": "UNET Loader Int8",
            "Quant8Model": "Model Quantization(int8)",
            "OneDiffQuantCheckpointLoaderSimple": "Load Checkpoint - OneDiff Quant",
            "OneDiffQuantCheckpointLoaderSimpleAdvanced": "Load Checkpoint - OneDiff Quant Advanced",
            "ImageOnlyOneDiffQuantCheckpointLoaderAdvanced": "Load Checkpoint - OneDiff Quant Advanced (img2vid)",
        }
    )



# import os
import dataclasses
import uuid
from abc import ABC, abstractmethod

# from functools import singledispatchmethod
# from typing import Optional

# import torch
# from comfy import model_management
# from comfy.controlnet import ControlLora, ControlNet
# from comfy.model_patcher import ModelPatcher
# from comfy.sd import VAE


class BoosterExecutor(ABC):
    """Interface for optimization."""

    @abstractmethod
    def execute(self, model, ckpt_name=None, **kwargs):
        """Apply the optimization strategy to the model."""
        pass


@dataclasses.dataclass
class BoosterSettings:
    tmp_cache_key: str = None


if __name__ == "__main__":
    print(BoosterSettings(str(uuid.uuid4())).tmp_cache_key)
    print(BoosterSettings(str(uuid.uuid4())).tmp_cache_key)
    print(BoosterSettings(str(uuid.uuid4())).tmp_cache_key)



import json
import os
from typing import NamedTuple

from core.registry import create_generator_registry
from core.service_client import ComfyGraph

WORKFLOW_DIR = "resources/workflows"
FACE_IMAGE_DIR = "/share_nfs/hf_models/comfyui_resources/input/faces"
POSE_IMAGE_DIR = "/share_nfs/hf_models/comfyui_resources/input/poses"
SDXL_MODELS = [
    "Juggernaut-XL_v9_RunDiffusionPhoto_v2.safetensors",
    "Pony_Realism.safetensors",
    "sdxl/dreamshaperXL_v21TurboDPMSDE.safetensors",
]
SD1_5_MODELS = [
    "sd15/020.realisticVisionV51_v51VAE.safetensors",
    "sd15/majicmixRealistic_v7.safetensors",
    "sd15/v1-5-pruned-emaonly.ckpt",
    "sd15/helloyoung25d_V10f.safetensors",
    "sd15/RealCartoonSpecialPruned.safetensors",
]


class InputParams(NamedTuple):
    graph: ComfyGraph


def read_prompts(file_path="resources/prompts.txt"):
    with open(file_path, "r", encoding="utf-8") as fp:
        lines = fp.readlines()
    return [line.strip() for line in lines if line.strip()]


def get_all_images(
    directory=FACE_IMAGE_DIR,
    image_extensions=set([".jpg", ".jpeg", ".png", ".gif", ".bmp"]),
):
    all_files = [os.path.join(directory, f) for f in os.listdir(directory)]
    image_files = [
        f
        for f in all_files
        if os.path.isfile(f) and os.path.splitext(f)[1].lower() in image_extensions
    ]
    return image_files


# Create register and get functions
register_generator, dispatch_generator = create_generator_registry()


@register_generator(f"{WORKFLOW_DIR}/example_workflow_api.json")
def _(workflow_path, *args, **kwargs):
    with open(workflow_path, "r") as fp:
        workflow = json.load(fp)
    graph = ComfyGraph(graph=workflow, sampler_nodes=["3"])
    for height in [1024, 768, 512]:
        for width in [1024, 768, 512]:
            graph.set_image_size(height=height, width=width)
            yield InputParams(graph=graph)


@register_generator(
    [f"{WORKFLOW_DIR}/baseline/txt2img.json", f"{WORKFLOW_DIR}/oneflow/txt2img.json"]
)
def _(workflow_path, *args, **kwargs):
    with open(workflow_path, "r") as fp:
        workflow = json.load(fp)
    graph = ComfyGraph(graph=workflow, sampler_nodes=["3"])
    for sdxl_model in SDXL_MODELS:
        graph.set_image_size(height=1024, width=1024)
        graph.graph["31"]["inputs"]["ckpt_name"] = sdxl_model
        yield InputParams(graph)
    for sd1_5_model in SD1_5_MODELS:
        graph.set_image_size(height=768, width=512)
        graph.graph["31"]["inputs"]["ckpt_name"] = sd1_5_model
        yield InputParams(graph)


SD3_WORKFLOWS = [
    f"{WORKFLOW_DIR}/baseline/sd3_basic.json",
    f"{WORKFLOW_DIR}/nexfort/sd3_basic.json",
]


@register_generator(SD3_WORKFLOWS)
def _(workflow_path, *args, **kwargs):
    with open(workflow_path, "r") as fp:
        workflow = json.load(fp)

    graph = ComfyGraph(graph=workflow, sampler_nodes=["271"])
    texts = read_prompts()
    for height in [1024, 768, 512]:
        for width in [1024, 768, 512]:
            for text in texts[-5:]:
                graph.set_prompt(prompt=text)
                graph.set_image_size(height=height, width=width)
                yield InputParams(graph=graph)


@register_generator(f"{WORKFLOW_DIR}/oneflow/sdxl-control-lora-speedup.json")
def _(workflow_path, *args, **kwargs):
    with open(workflow_path, "r") as fp:
        workflow = json.load(fp)

    graph = ComfyGraph(graph=workflow, sampler_nodes=["1"])
    yield InputParams(graph=graph)


@register_generator(
    [
        f"{WORKFLOW_DIR}/baseline/ComfyUI_IPAdapter_plus/ipadapter_advanced.json",
        f"{WORKFLOW_DIR}/oneflow/ComfyUI_IPAdapter_plus/ipadapter_advanced.json",
    ]
)
def _(workflow_path, *args, **kwargs):
    with open(workflow_path, "r") as fp:
        workflow = json.load(fp)

    graph = ComfyGraph(graph=workflow, sampler_nodes=["3"])
    for image in get_all_images():
        graph.graph["12"]["inputs"]["image"] = image
        for height in [768, 512]:
            for width in [768, 512]:
                positive_prompt = "in a peaceful spring morning a woman wearing a white shirt is sitting in a park on a bench\n\nhigh quality, detailed, diffuse light"
                negative_prompt = "blurry, noisy, messy, lowres, jpeg, artifacts, ill, distorted, malformed"
                graph.set_prompt(positive_prompt, negative_prompt)
                graph.set_image_size(height=height, width=width)
                yield InputParams(graph=graph)


@register_generator(
    [
        f"{WORKFLOW_DIR}/baseline/lora.json",
        f"{WORKFLOW_DIR}/baseline/lora_multiple.json",
        f"{WORKFLOW_DIR}/oneflow/lora_speedup.json",
        f"{WORKFLOW_DIR}/oneflow/lora_multiple_speedup.json",
    ]
)
def _(workflow_path, *args, **kwargs):
    with open(workflow_path, "r") as fp:
        workflow = json.load(fp)

    graph = ComfyGraph(graph=workflow, sampler_nodes=["3"])
    graph.set_prompt("masterpiece best quality girl, hanfu", "bad hands")
    root_path = "sd15/"
    checkpoint_nodes = []
    for node in graph.graph.values():
        if node["class_type"] in [
            "CheckpointLoaderSimple",
            "OneDiffCheckpointLoaderSimple",
        ]:
            checkpoint_nodes.append(node)
    assert len(checkpoint_nodes) == 1

    for file_name in [
        "020.realisticVisionV51_v51VAE.safetensors",
        "v1-5-pruned-emaonly.ckpt",
    ]:
        checkpoint_path = os.path.join(root_path, file_name)
        checkpoint_nodes[0]["inputs"]["ckpt_name"] = checkpoint_path
        yield InputParams(graph=graph)


@register_generator(
    [
        f"{WORKFLOW_DIR}/baseline/ComfyUI_InstantID/instantid_posed.json",
        f"{WORKFLOW_DIR}/oneflow/ComfyUI_InstantID/instantid_posed_speedup.json",
    ]
)
def _(workflow_path, *args, **kwargs):
    with open(workflow_path, "r") as fp:
        workflow = json.load(fp)
    graph = ComfyGraph(graph=workflow, sampler_nodes=["3"])

    face_imgs = get_all_images(FACE_IMAGE_DIR)[:2]
    pose_imgs = get_all_images(POSE_IMAGE_DIR)[:2]
    for face_img in face_imgs:
        for pose_img in pose_imgs:
            # print(f'{face_img=} {pose_img=}')
            graph.graph["13"]["inputs"]["image"] = face_img
            graph.graph["67"]["inputs"]["image"] = pose_img
            yield InputParams(graph=graph)


@register_generator(
    [
        f"{WORKFLOW_DIR}/baseline/PuLID_ComfyUI/PuLID_4-Step_lightning.json",
        f"{WORKFLOW_DIR}/oneflow/PuLID_ComfyUI/PuLID_4-Step_lightning.json",
    ]
)
def _(workflow_path, *args, **kwargs):
    with open(workflow_path, "r") as fp:
        workflow = json.load(fp)
    graph = ComfyGraph(graph=workflow, sampler_nodes=["3"])

    for height in [768, 512]:
        for width in [768]:
            for face_img in get_all_images(FACE_IMAGE_DIR):
                graph.graph["12"]["inputs"]["image"] = face_img
                graph.set_image_size(height=height, width=width)
                yield InputParams(graph=graph)


@register_generator(
    [
        f"{WORKFLOW_DIR}/baseline/PuLID_ComfyUI/PuLID_IPAdapter_style_transfer.json",
        f"{WORKFLOW_DIR}/oneflow/PuLID_ComfyUI/PuLID_IPAdapter_style_transfer.json",
    ]
)
def _(workflow_path, *args, **kwargs):
    with open(workflow_path, "r") as fp:
        workflow = json.load(fp)
    graph = ComfyGraph(graph=workflow, sampler_nodes=["3"])
    face_imgs = get_all_images(FACE_IMAGE_DIR)[:2]
    pose_imgs = get_all_images(POSE_IMAGE_DIR)[:2]
    for face_img in face_imgs:
        for pose_img in pose_imgs:
            # print(f'{face_img=} {pose_img=}')
            graph.graph["12"]["inputs"]["image"] = face_img
            graph.graph["48"]["inputs"]["image"] = pose_img
            yield InputParams(graph=graph)



"""OneDiff ComfyUI Speedup Module"""
from onediff.utils.import_utils import is_nexfort_available, is_oneflow_available

from ._config import is_disable_oneflow_backend
from ._nodes import (
    ControlnetSpeedup,
    ModelSpeedup,
    OneDiffApplyModelBooster,
    OneDiffCheckpointLoaderSimple,
    OneDiffControlNetLoader,
    VaeSpeedup,
)

NODE_CLASS_MAPPINGS = {
    "ModelSpeedup": ModelSpeedup,
    "VaeSpeedup": VaeSpeedup,
    "ControlnetSpeedup": ControlnetSpeedup,
    "OneDiffModelBooster": OneDiffApplyModelBooster,
    "OneDiffCheckpointLoaderSimple": OneDiffCheckpointLoaderSimple,
    "OneDiffControlNetLoader": OneDiffControlNetLoader,
}

NODE_DISPLAY_NAME_MAPPINGS = {
    "ModelSpeedup": "Model Speedup",
    "VaeSpeedup": "VAE Speedup",
    "OneDiffModelBooster": "Apply Model Booster - OneDiff",
    "ControlnetSpeedup": "ControlNet Speedup",
    "OneDiffCheckpointLoaderSimple": "Load Checkpoint - OneDiff",
}


def update_node_mappings(node):
    NODE_CLASS_MAPPINGS.update(node.NODE_CLASS_MAPPINGS)
    NODE_DISPLAY_NAME_MAPPINGS.update(node.NODE_DISPLAY_NAME_MAPPINGS)


def lazy_load_extra_nodes():

    from .extras_nodes import nodes_torch_compile_booster

    update_node_mappings(nodes_torch_compile_booster)

    if is_oneflow_available() and not is_disable_oneflow_backend():
        from .extras_nodes import nodes_compare, nodes_oneflow_booster

        update_node_mappings(nodes_oneflow_booster)
        update_node_mappings(nodes_compare)

    if is_nexfort_available():
        from .extras_nodes import nodes_nexfort_booster

        update_node_mappings(nodes_nexfort_booster)

    from .extras_nodes import nodes_prompt_styler

    update_node_mappings(nodes_prompt_styler)


# Lazy load all extra nodes when needed
lazy_load_extra_nodes()



# Compile and save to oneflow graph example: python examples/text_to_image_sdxl_mp_load.py --save
# Compile and load to new device example: python examples/text_to_image_sdxl_mp_load.py --load

import argparse
import os

import torch
import oneflow as flow  # usort: skip

parser = argparse.ArgumentParser()
parser.add_argument(
    "--base", type=str, default="stabilityai/stable-diffusion-xl-base-1.0"
)
parser.add_argument("--variant", type=str, default="fp16")
parser.add_argument(
    "--prompt",
    type=str,
    default="street style, detailed, raw photo, woman, face, shot on CineStill 800T",
)
parser.add_argument("--n_steps", type=int, default=30)
parser.add_argument("--saved_image", type=str, required=False, default="sdxl-out.png")
parser.add_argument("--seed", type=int, default=1)
parser.add_argument("--save", action=argparse.BooleanOptionalAction)
parser.add_argument("--load", action=argparse.BooleanOptionalAction)
parser.add_argument("--file", type=str, required=False, default="unet_compiled")
args = parser.parse_args()


def run_sd(cmd_args, device):
    from diffusers import DiffusionPipeline
    from onediff.infer_compiler import oneflow_compile

    # Normal SDXL pipeline init.
    seed = torch.Generator(device).manual_seed(cmd_args.seed)
    output_type = "pil"
    # SDXL base: StableDiffusionXLPipeline
    base = DiffusionPipeline.from_pretrained(
        cmd_args.base,
        torch_dtype=torch.float16,
        variant=cmd_args.variant,
        use_safetensors=True,
    )
    base.to(device)

    # Compile unet with oneflow
    print("unet is compiled to oneflow.")
    base.unet = oneflow_compile(base.unet)

    # Load compiled unet with oneflow
    if cmd_args.load:
        print("loading graphs...")
        base.unet.load_graph("base_" + cmd_args.file, device)

    # Normal SDXL run
    # sizes = [1024, 896, 768]
    sizes = [1024]
    for h in sizes:
        for w in sizes:
            for i in range(1):
                image = base(
                    prompt=cmd_args.prompt,
                    height=h,
                    width=w,
                    generator=seed,
                    num_inference_steps=cmd_args.n_steps,
                    output_type=output_type,
                ).images
                image[0].save(f"h{h}-w{w}-i{i}-{cmd_args.saved_image}")

    # Save compiled unet with oneflow
    if cmd_args.save:
        print("saving graphs...")
        base.unet.save_graph("base_" + cmd_args.file)


if __name__ == "__main__":
    if args.save:
        run_sd(args, "cuda:0")

    if args.load:
        import torch.multiprocessing as mp

        # multi device/process run
        devices = ("cuda:0", "cuda:1")
        procs = []
        for device in devices:
            p = mp.get_context("spawn").Process(target=run_sd, args=(args, device))
            p.start()
            procs.append(p)

        for p in procs:
            p.join()
            print(p)



from abc import ABC, abstractmethod
from typing import Dict, List

from comfy.model_base import BaseModel

from comfy.model_patcher import ModelPatcher


class PatchExecutorBase(ABC):
    @abstractmethod
    def check_patch(self):
        pass

    @abstractmethod
    def set_patch(self):
        pass

    @abstractmethod
    def get_patch(self):
        pass


class UiNodeWithIndexPatch(PatchExecutorBase):
    DEFAULT_VALUE = -1
    INCREMENT_VALUE = 1

    def __init__(self) -> None:
        self.patch_name = type(self).__name__

    def check_patch(self, module: ModelPatcher) -> bool:
        return hasattr(module, self.patch_name)

    def set_patch(self, module: ModelPatcher, value: int):
        setattr(module, self.patch_name, value)

    def get_patch(self, module: ModelPatcher) -> int:
        return getattr(module, self.patch_name, self.DEFAULT_VALUE)

    def copy_to(self, old_model: ModelPatcher, new_model: ModelPatcher):
        value = self.get_patch(old_model)
        self.set_patch(new_model, value + self.INCREMENT_VALUE)


class CachedCrossAttentionPatch(PatchExecutorBase):
    def __init__(self) -> None:
        self.patch_name = type(self).__name__

    def check_patch(self, module):
        return hasattr(module, self.patch_name)

    def set_patch(self, module, value: dict):
        setattr(module, self.patch_name, value)

    def get_patch(self, module) -> Dict[str, any]:
        if not self.check_patch(module):
            self.set_patch(module, {})
        return getattr(module, self.patch_name)

    def clear_patch(self, module):
        if self.check_patch(module):
            self.get_patch(module).clear()


class CrossAttentionForwardMasksPatch(PatchExecutorBase):
    def __init__(self) -> None:
        """Will be abandoned"""
        self.patch_name = "forward_masks"

    def check_patch(self, module):
        return hasattr(module, self.patch_name)

    def set_patch(self, module, value):
        raise NotImplementedError()

    def get_patch(self, module) -> Dict:
        if not self.check_patch(module):
            setattr(module, self.patch_name, {})
        return getattr(module, self.patch_name)

    def clear_patch(self, module):
        if self.check_patch(module):
            self.get_patch(module).clear()


class DeepCacheUNetExecutorPatch(PatchExecutorBase):
    def __init__(self) -> None:
        super().__init__()
        self.patch_names = ("deep_cache_unet", "fast_deep_cache_unet")

    def check_patch(self, model_patcher: ModelPatcher):
        return all(hasattr(model_patcher, name) for name in self.patch_names)

    def set_patch(self, model_patcher: ModelPatcher, values):
        assert len(self.patch_names) == len(values)
        for attr, value in zip(self.patch_names, values):
            setattr(model_patcher, attr, value)

    def get_patch(self, model_patcher: ModelPatcher) -> List:
        return [getattr(model_patcher, attr, None) for attr in self.patch_names]

    def copy_to(self, old_model: ModelPatcher, new_model: ModelPatcher):
        values = self.get_patch(old_model)
        self.set_patch(new_model, values)
        new_model.model.use_deep_cache_unet = True

    def is_use_deep_cache_unet(self, module: BaseModel):
        return getattr(module, "use_deep_cache_unet", False)


class UNetExtraInputOptions(PatchExecutorBase):
    def __init__(self) -> None:
        """UNetExtraInputOptions"""
        super().__init__()
        self.patch_name = type(self).__name__

    def check_patch(self, module):
        return hasattr(module, self.patch_name)

    def set_patch(self, module, value: Dict):
        """
        Bind extra input options to the specified module.
        For UNet extra input options, the value is a dictionary.

        Args:
            module: The module object to set the patch attribute on.
            value (Dict): The extra input options to bind to the module.
        """
        setattr(module, self.patch_name, value)

    def get_patch(self, module) -> Dict:
        if not self.check_patch(module):
            self.set_patch(module, {})
        return getattr(module, self.patch_name)

    def clear_patch(self, module):
        if self.check_patch(module):
            self.get_patch(module).clear()



import argparse
import importlib.metadata

from diffusers import DiffusionPipeline
from packaging import version


def check_diffusers_version():
    required_version = version.parse("0.22.0")
    package_name = "diffusers"

    try:
        installed_version = version.parse(importlib.metadata.version(package_name))
        if installed_version < required_version:
            raise ValueError(
                f"Installed {package_name} version ({installed_version}) is lower than required ({required_version})"
            )

    except importlib.metadata.PackageNotFoundError:
        print(f"{package_name} is not installed")


def parse_args():
    parser = argparse.ArgumentParser(description="Simple demo of LCM image generation.")
    parser.add_argument(
        "--prompt",
        type=str,
        default="Self-portrait oil painting, a beautiful cyborg with golden hair, 8k",
    )
    parser.add_argument(
        "--model_id",
        type=str,
        default="SimianLuo/LCM_Dreamshaper_v7",
        help="Model id or local path to the LCM model.",
    )
    parser.add_argument("--height", type=int, default=768)
    parser.add_argument("--width", type=int, default=768)
    parser.add_argument("--steps", type=int, default=4)
    parser.add_argument(
        "--warmup",
        type=int,
        default=1,
        help="Warmup option sets the number of preliminary runs of the pipeline. These initial runs are necessary to ensure accurate performance data during testing.",
    )
    parser.add_argument("--seed", type=int, default=1)
    parser.add_argument(
        "--disable", action="store_true", help="Disable Onediff speeding up."
    )
    args = parser.parse_args()
    return args


check_diffusers_version()

args = parse_args()

if not args.disable:
    from onediff.infer_compiler import oneflow_compile
import torch

pipe = DiffusionPipeline.from_pretrained(args.model_id)

pipe.to(torch_device="cuda", torch_dtype=torch.float16)
if not args.disable:
    pipe.unet = oneflow_compile(pipe.unet)

for _ in range(args.warmup):
    torch.manual_seed(args.seed)
    images = pipe(
        args.prompt,
        height=args.height,
        width=args.width,
        num_inference_steps=args.steps,
    ).images

torch.manual_seed(args.seed)
images = pipe(
    args.prompt, height=args.height, width=args.width, num_inference_steps=args.steps
).images
for i, image in enumerate(images):
    image.save(
        f"LCM-{args.width}x{args.height}-seed-{args.seed}-disable-{args.disable}.png"
    )



import os
import sys

import folder_paths
import torch

__all__ = [
    "is_default_using_oneflow_backend",
    "is_default_using_nexfort_backend",
    "is_disable_oneflow_backend",
]

# https://github.com/comfyanonymous/ComfyUI/blob/master/folder_paths.py#L9
os.environ["COMFYUI_ROOT"] = folder_paths.base_path
_default_backend = os.environ.get("ONEDIFF_COMFY_NODES_DEFAULT_BACKEND", "oneflow")
_disable_oneflow_backend = (
    os.environ.get("ONEDIFF_COMFY_NODES_DISABLE_ONEFLOW_BACKEND", "0") == "1"
)

custom_nodes_path = os.path.join(folder_paths.base_path, "custom_nodes")

# Add paths to sys.path if not already there
if custom_nodes_path not in sys.path:
    sys.path.append(custom_nodes_path)


if _default_backend not in ["oneflow", "nexfort"]:
    raise ValueError(f"Invalid default backend: {_default_backend}")


def is_disable_oneflow_backend():
    return _disable_oneflow_backend


def is_default_using_nexfort_backend():
    return _default_backend == "nexfort"


def is_default_using_oneflow_backend():
    return _default_backend == "oneflow" and not _disable_oneflow_backend



import oneflow as flow  # usort: skip
from onediff.infer_compiler.backends.oneflow.transform import proxy_class
from sgm.modules.attention import CrossAttention, SpatialTransformer
from sgm.modules.diffusionmodules.openaimodel import UNetModel
from sgm.modules.diffusionmodules.util import GroupNorm32

from .common import CrossAttentionOflow, GroupNorm32Oflow, timestep_embedding


def cat(tensors, *args, **kwargs):
    if len(tensors) == 2:
        a, b = tensors
        a = flow.nn.functional.interpolate_like(a, like=b, mode="nearest")
        tensors = (a, b)
    return flow.cat(tensors, *args, **kwargs)


# https://github.com/Stability-AI/generative-models/blob/059d8e9cd9c55aea1ef2ece39abf605efb8b7cc9/sgm/modules/diffusionmodules/openaimodel.py#L816
class UNetModelOflow(proxy_class(UNetModel)):
    def forward(self, x, timesteps=None, context=None, y=None, **kwargs):
        assert (y is not None) == (
            self.num_classes is not None
        ), "must specify y if and only if the model is class-conditional"
        hs = []
        t_emb = timestep_embedding(timesteps, self.model_channels).half()
        emb = self.time_embed(t_emb)
        x = x.half()
        context = context.half() if context is not None else context
        y = y.half() if y is not None else y
        if self.num_classes is not None:
            assert y.shape[0] == x.shape[0]
            emb = emb + self.label_emb(y)
        h = x
        for module in self.input_blocks:
            h = module(h, emb, context)
            hs.append(h)
        h = self.middle_block(h, emb, context)
        for module in self.output_blocks:
            # h = flow.cat([h, hs.pop()], dim=1)
            h = cat([h, hs.pop()], dim=1)
            h = module(h, emb, context)
        h = h.type(x.dtype)
        return self.out(h)


class SpatialTransformerOflow(proxy_class(SpatialTransformer)):
    # https://github.com/Stability-AI/generative-models/blob/059d8e9cd9c55aea1ef2ece39abf605efb8b7cc9/sgm/modules/attention.py#L702
    def forward(self, x, context=None):
        # note: if no context is given, cross-attention defaults to self-attention
        if not isinstance(context, list):
            context = [context]
        b, c, h, w = x.shape
        x_in = x
        x = self.norm(x)
        if not self.use_linear:
            x = self.proj_in(x)
        x = x.flatten(2, 3).permute(0, 2, 1)
        if self.use_linear:
            x = self.proj_in(x)
        for i, block in enumerate(self.transformer_blocks):
            if i > 0 and len(context) == 1:
                i = 0  # use same context for each block
            x = block(x, context=context[i])
        if self.use_linear:
            x = self.proj_out(x)
        x = x.permute(0, 2, 1).reshape_as(x_in)
        if not self.use_linear:
            x = self.proj_out(x)
        return x + x_in


torch2oflow_class_map = {
    CrossAttention: CrossAttentionOflow,
    GroupNorm32: GroupNorm32Oflow,
    SpatialTransformer: SpatialTransformerOflow,
    UNetModel: UNetModelOflow,
}



"""[SDXL](https://huggingface.co/stabilityai/stable-diffusion-xl-base-1.0)

## Performance Comparison

Updated on Mon 08 Apr 2024

Timings for 30 steps at 1024x1024
| Accelerator             | Baseline (non-optimized) | OneDiff(optimized) | OneDiff Quant(optimized) |
| ----------------------- | ------------------------ | ------------------ | ------------------------ |
| NVIDIA GeForce RTX 3090 | 8.03 s                   | 4.44 s ( ~44.7%)   | 3.34 s ( ~58.4%)         |

- torch   {version: 2.2.1+cu121}
- oneflow {git_commit: 710818c, version: 0.9.1.dev20240406+cu121, enterprise: True}

## Install

1. [OneDiff Installation Guide](https://github.com/siliconflow/onediff/blob/main/README_ENTERPRISE.md#install-onediff-enterprise)
2. [OneDiffx Installation Guide](https://github.com/siliconflow/onediff/tree/main/onediff_diffusers_extensions#install-and-setup)

## Usage:
> onediff/onediff_diffusers_extensions/examples/text_to_image_online_quant.py

```shell
# Baseline (non-optimized)
$   python text_to_image_online_quant.py \
        --model_id  /share_nfs/hf_models/stable-diffusion-xl-base-1.0  \
        --seed 1 \
        --backend torch  --height 1024 --width 1024 --output_file sdxl_torch.png
```
```shell
# OneDiff Quant(optimized)
$   python text_to_image_online_quant.py \
        --model_id  /share_nfs/hf_models/stable-diffusion-xl-base-1.0  \
        --seed 1 \
        --backend onediff \
        --cache_dir ./run_sdxl_quant \
        --height 1024 \
        --width 1024 \
        --output_file sdxl_quant.png   \
        --quantize \
        --conv_mae_threshold 0.1 \
        --linear_mae_threshold 0.2 \
        --conv_compute_density_threshold 900 \
        --linear_compute_density_threshold 300
```

| Option                                 | Range  | Default | Description                                                                  |
| -------------------------------------- | ------ | ------- | ---------------------------------------------------------------------------- |
| --conv_mae_threshold 0.1               | [0, 1] | 0.1     | MAE threshold for quantizing convolutional modules to 0.1.                   |
| --linear_mae_threshold 0.2             | [0, 1] | 0.2     | MAE threshold for quantizing linear modules to 0.2.                          |
| --conv_compute_density_threshold 900   | [0, ∞) | 900     | Computational density threshold for quantizing convolutional modules to 900. |
| --linear_compute_density_threshold 300 | [0, ∞) | 300     | Computational density threshold for quantizing linear modules to 300.        |

Notes:

1. Set CUDA device using export CUDA_VISIBLE_DEVICES=7.

2. The log *.pt file is cached. Quantization result information can be found in `cache_dir`/quantization_stats.json.

"""
import argparse
import time

import torch
from diffusers import AutoPipelineForText2Image
from onediff_quant.quantization import QuantizationConfig
from onediffx import compile_pipe


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--model_id", default="runwayml/stable-diffusion-v1-5")
    parser.add_argument(
        "--prompt", default="a photo of an astronaut riding a horse on mars"
    )
    parser.add_argument(
        "--output_file", default="astronaut_rides_horse_onediff_quant.png"
    )
    parser.add_argument("--seed", type=int, default=1)
    parser.add_argument("--backend", default="onediff", choices=["onediff", "torch"])
    parser.add_argument("--quantize", action="store_true")
    parser.add_argument("--cache_dir", default="./run_sd-v1-5")
    parser.add_argument("--height", type=int, default=1024)
    parser.add_argument("--width", type=int, default=1024)
    parser.add_argument("--num_inference_steps", type=int, default=30)
    parser.add_argument("--conv_mae_threshold", type=float, default=0.2)
    parser.add_argument("--linear_mae_threshold", type=float, default=0.4)
    parser.add_argument("--conv_compute_density_threshold", type=int, default=900)
    parser.add_argument("--linear_compute_density_threshold", type=int, default=300)
    return parser.parse_args()


def load_model(model_id):
    pipe = AutoPipelineForText2Image.from_pretrained(
        model_id, torch_dtype=torch.float16, variant="fp16"
    )
    pipe.to(f"cuda")
    return pipe


def compile_and_quantize_model(pipe, cache_dir, quantize, quant_params):
    pipe = compile_pipe(pipe)
    if quantize:
        config = QuantizationConfig.from_settings(
            **quant_params, cache_dir=cache_dir, plot_calibrate_info=True
        )
        pipe.unet.apply_online_quant(quant_config=config)
    return pipe


def save_image(image, output_file):
    image.save(output_file)
    print(f"Image saved to: {output_file}")


def main():
    args = parse_args()
    pipe = load_model(args.model_id)
    if args.backend == "onediff":
        compile_and_quantize_model(
            pipe,
            args.cache_dir,
            args.quantize,
            {
                "conv_mae_threshold": args.conv_mae_threshold,
                "linear_mae_threshold": args.linear_mae_threshold,
                "conv_compute_density_threshold": args.conv_compute_density_threshold,
                "linear_compute_density_threshold": args.linear_compute_density_threshold,
            },
        )
    torch.manual_seed(args.seed)
    # Warm-up
    pipe(prompt=args.prompt, num_inference_steps=1)

    # Run_inference
    for _ in range(5):
        start_time = time.time()
        torch.manual_seed(args.seed)
        image = pipe(
            prompt=args.prompt,
            height=args.height,
            width=args.width,
            num_inference_steps=args.num_inference_steps,
        ).images[0]
        end_time = time.time()
        print(f"Inference time: {end_time - start_time:.2f} seconds")

    save_image(image, args.output_file)


if __name__ == "__main__":
    main()



from ..booster_utils import is_using_nexfort_backend
from ..hijack_ipadapter_plus.set_model_patch_replace import set_model_patch_replace
from ._config import comfyui_instantid, comfyui_instantid_hijacker

set_model_patch_replace_fn_pt = comfyui_instantid.InstantID._set_model_patch_replace


def cond_func(org_fn, model, *args, **kwargs):
    return is_using_nexfort_backend(model)


comfyui_instantid_hijacker.register(
    set_model_patch_replace_fn_pt, set_model_patch_replace, cond_func
)



import logging

logger = logging.getLogger("onediff-benchmark")
logger.setLevel(logging.INFO)
formatter = logging.Formatter(
    "%(asctime)s - %(levelname)s - %(message)s", "%Y-%m-%d %H:%M:%S"
)
console_handler = logging.StreamHandler()
console_handler.setFormatter(formatter)
logger.addHandler(console_handler)



from functools import partial, singledispatchmethod
from typing import Optional

import torch
from comfy.controlnet import ControlLora, ControlNet
from comfy.model_patcher import ModelPatcher
from comfy.sd import VAE
from onediff.infer_compiler.backends.nexfort.deployable_module import (
    get_deployable_module,
)

from ..booster_interface import BoosterExecutor


def compile(model: callable, *args, **kwargs):
    compiled_model = torch.compile(model, *args, **kwargs)
    return get_deployable_module(model, compiled_model)


class TorchCompileBoosterExecutor(BoosterExecutor):
    # https://pytorch.org/docs/stable/_modules/torch.html#compile
    def __init__(
        self,
        fullgraph=False,
        dynamic=None,
        backend="inductor",
        mode="default",
        options=None,
        disable=False,
    ):
        super().__init__()
        self.compile_kwargs = {
            "fullgraph": fullgraph,
            "dynamic": dynamic,
            "backend": backend,
            "options": options,
            "mode": mode,
            "disable": disable,
        }
        self.compile_fn = partial(compile, **self.compile_kwargs)

    @singledispatchmethod
    def execute(self, model, ckpt_name=None, **kwargs):
        raise NotImplementedError(f"Cannot execute {type(model)=}")

    @execute.register(ModelPatcher)
    def _(self, model, ckpt_name: Optional[str] = None, **kwargs):
        model.model.diffusion_model.to(memory_format=torch.channels_last)
        model.model.diffusion_model = self.compile_fn(model.model.diffusion_model)
        return model

    @execute.register(VAE)
    def _(self, model, ckpt_name: Optional[str] = None, **kwargs):
        # https://huggingface.co/blog/sd3#performance-optimizations-for-sd3
        model.first_stage_model.to(memory_format=torch.channels_last)
        model.first_stage_model.decode = self.compile_fn(model.first_stage_model.decode)
        return model

    @execute.register(ControlNet)
    def _(self, model, ckpt_name: Optional[str] = None, **kwargs):
        torch_model = model.control_model
        compiled_model = self.compile_fn(torch_model)
        model.control_model = compiled_model
        return model



import os

from modules import shared


class SD21CompileCtx(object):
    """to avoid results for NaN when the model is v2-1_768-ema-pruned"""

    _var_name = "ONEFLOW_ATTENTION_ALLOW_HALF_PRECISION_ACCUMULATION"

    def __enter__(self):
        self._original = os.getenv(self._var_name)
        if shared.opts.sd_model_checkpoint.startswith("v2-1"):
            os.environ[self._var_name] = "0"

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._original is not None:
            os.environ[self._var_name] = self._original
        return False



import argparse
import os
import time

import torch
import torch.nn as nn

from onediff.infer_compiler import oneflow_compile, OneflowCompileOptions


def parse_args():
    parser = argparse.ArgumentParser()

    parser.add_argument("--model", type=str, required=True)
    parser.add_argument("--saved_image", type=str, required=True)
    parser.add_argument("--save_graph", action="store_true")
    parser.add_argument("--load_graph", action="store_true")
    parser.add_argument(
        "--prompt",
        type=str,
        default="a photo of an astronaut riding a horse on mars",
    )
    parser.add_argument("--height", type=int, default=512)
    parser.add_argument("--width", type=int, default=512)
    parser.add_argument("--steps", type=int, default=30)
    parser.add_argument("--bits", type=int, default=8)
    parser.add_argument(
        "--compile",
        default=True,
        type=(lambda x: str(x).lower() in ["true", "1", "yes"]),
    )
    parser.add_argument(
        "--compile_text_encoder",
        default=False,
        type=(lambda x: str(x).lower() in ["true", "1", "yes"]),
        help=(
            "Switch controls whether text_encoder is compiled (default: False). "
            "If your CPU is powerful, turning it on will shorten end-to-end time."
        ),
    )
    parser.add_argument(
        "--graph",
        default=True,
        type=(lambda x: str(x).lower() in ["true", "1", "yes"]),
    )
    parser.add_argument("--seed", type=int, default=1)
    parser.add_argument("--warmup", type=int, default=1)

    args = parser.parse_args()
    return args


args = parse_args()

assert os.path.isfile(
    os.path.join(args.model, "calibrate_info.txt")
), f"calibrate_info.txt is required in args.model ({args.model})"

import onediff_quant
from diffusers import StableDiffusionPipeline
from onediff_quant.utils import replace_sub_module_with_quantizable_module

onediff_quant.enable_load_quantized_model()
infer_args = {
    "prompt": args.prompt,
    "height": args.height,
    "width": args.width,
    "num_inference_steps": args.steps,
}
calibrate_info = {}
with open(os.path.join(args.model, "calibrate_info.txt"), "r") as f:
    for line in f.readlines():
        line = line.strip()
        items = line.split(" ")
        calibrate_info[items[0]] = [
            float(items[1]),
            int(items[2]),
            [float(x) for x in items[3].split(",")],
        ]

pipe = StableDiffusionPipeline.from_pretrained(
    args.model,
    use_auth_token=True,
    revision="fp16",
    variant="fp16",
    torch_dtype=torch.float16,
    safety_checker=None,
)
pipe.to("cuda")


for sub_module_name, sub_calibrate_info in calibrate_info.items():
    replace_sub_module_with_quantizable_module(
        pipe.unet,
        sub_module_name,
        sub_calibrate_info,
        False,
        False,
        args.bits,
    )

compile_options = OneflowCompileOptions()
compile_options.use_graph = args.graph

if args.compile_text_encoder:
    if pipe.text_encoder is not None:
        pipe.text_encoder = oneflow_compile(pipe.text_encoder, options=compile_options)

if args.compile:
    pipe.unet = oneflow_compile(pipe.unet, options=compile_options)
    pipe.vae.decoder = oneflow_compile(pipe.vae.decoder, options=compile_options)


if args.load_graph:
    print("Loading graphs to avoid compilation...")
    start_t = time.time()
    pipe.unet.load_graph("base_unet_compiled", run_warmup=True)
    pipe.vae.decoder.load_graph("base_vae_compiled", run_warmup=True)
    end_t = time.time()
    print(f"warmup with loading graph elapsed: {end_t - start_t} s")
    start_t = time.time()
    for _ in range(args.warmup):
        torch.manual_seed(args.seed)
        image = pipe(**infer_args).images[0]
    end_t = time.time()
    print(f"warmup with run elapsed: {end_t - start_t} s")
else:
    start_t = time.time()
    for _ in range(args.warmup):
        torch.manual_seed(args.seed)
        image = pipe(**infer_args).images[0]
    end_t = time.time()
    print(f"warmup with run elapsed: {end_t - start_t} s")

start_t = time.time()

torch.manual_seed(args.seed)
torch.cuda.cudart().cudaProfilerStart()
image = pipe(**infer_args).images[0]
torch.cuda.cudart().cudaProfilerStop()

end_t = time.time()
print(f"e2e ({args.steps} steps) elapsed: {end_t - start_t} s")

image.save(args.saved_image)

if args.save_graph:
    print("Saving graphs...")
    start_t = time.time()
    pipe.unet.save_graph("base_unet_compiled")
    pipe.vae.decoder.save_graph("base_vae_compiled")
    end_t = time.time()
    print(f"save graphs elapsed: {end_t - start_t} s")



import os
import random
import subprocess

import folder_paths
import numpy as np
import oneflow as flow  # usort: skip
from onediff.infer_compiler.backends.oneflow.transform.builtin_transform import (
    torch2oflow,
)
from PIL import Image

try:
    from skimage.metrics import structural_similarity
except ImportError as e:
    subprocess.check_call(["pip", "install", "scikit-image"])
    from skimage.metrics import structural_similarity


class CompareModel:
    r"""
    This node is used to compare the weight values corresponding to each same key in PyTorch and Oneflow Graph,
    and to verify the correctness of Oneflow Graph compilation, especially for models with LoRA.

    To use it, you MUST load two same checkpoints (one for PyTorch and another for OneFlow Graph)
    and connect them to the two inputs of the node, then set `check` to enable.

    If there is too much difference between their weights, the node will raise RuntimeError.

    Note:
        The first time you run workflow with OneFlow Graph, the node cannot capture the weights of OneFlow Graph without compiling,
        so you need to run workflow at least twice to ensure that this node works properly.
    """

    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "torch_model": ("MODEL",),
                "oneflow_model": ("MODEL",),
                "check": (["enable", "disable"],),
            },
        }

    RETURN_TYPES = ()
    RETURN_NAMES = ()
    OUTPUT_NODE = True
    FUNCTION = "compare"
    CATEGORY = "OneDiff"

    def compare(self, torch_model, oneflow_model, check):
        if check == "disable":
            return {}

        torch_model.patch_model("cuda")
        oneflow_model.patch_model("cuda")

        torch_unet = torch_model.model.diffusion_model
        oflow_unet = oneflow_model.model.diffusion_model

        if oflow_unet._deployable_module_model._oneflow_module is None:
            torch_model.unpatch_model("cuda")
            oneflow_model.unpatch_model("cuda")
            print(
                "\033[91m",
                f"[Compare Model Diff] Cannot get compiled oneflow graph, maybe you need to build it"
                + "\033[0m",
            )
            return {}

        removeprefix = (
            lambda ss, prefix: ss[len(prefix) :] if ss.startswith(prefix) else ss
        )

        cnt = 0
        for key, _ in oflow_unet.named_parameters():
            key = removeprefix(key, "_deployable_module_model._torch_module.")
            torch_value = torch_unet.get_parameter(key).cuda()
            oflow_value = (
                oflow_unet._deployable_module_model._oneflow_module.get_parameter(
                    key
                ).cuda()
            )

            if not flow.allclose(torch2oflow(torch_value), oflow_value, 1e-4, 1e-4):
                print(
                    "\033[91m",
                    f"value of key [{key}] is different, max diff is {(torch2oflow(torch_value) - oflow_value).float().abs().max().item()}"
                    + "\033[0m",
                )
                cnt += 1

            if cnt >= 10:
                torch_model.unpatch_model("cuda")
                oneflow_model.unpatch_model("cuda")
                raise RuntimeError(
                    "Too much lora weight diff between torch and oneflow, omit..."
                )

        torch_model.unpatch_model("cuda")
        oneflow_model.unpatch_model("cuda")
        return {}


class ShowImageDiff:
    def __init__(self):
        self.output_dir = folder_paths.get_temp_directory()
        self.type = "temp"
        self.prefix_append = "_temp_" + "".join(
            random.choice("abcdefghijklmnopqrstupvxyz") for x in range(5)
        )

    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "images1": ("IMAGE",),
                "images2": ("IMAGE",),
                "ssim_threshold": ("STRING", {"default": "0.9"}),
                "raise_if_diff": (["enable", "disable"],),
                "image_id": ("STRING", {"default": "ComfyUI"}),
            },
            "hidden": {"prompt": "PROMPT", "extra_pnginfo": "EXTRA_PNGINFO"},
        }

    RETURN_TYPES = ()
    FUNCTION = "save_images"

    OUTPUT_NODE = True

    CATEGORY = "OneDiff"

    def save_images(
        self,
        images1,
        images2,
        ssim_threshold: float,
        raise_if_diff: str,
        image_id="ComfyUI",
        prompt=None,
        extra_pnginfo=None,
    ):
        assert len(images1) == len(images2)
        filename_prefix = image_id[:]
        filename_prefix += self.prefix_append
        (
            full_output_folder,
            filename,
            counter,
            subfolder,
            filename_prefix,
        ) = folder_paths.get_save_image_path(
            filename_prefix,
            self.output_dir,
            images1[0].shape[1],
            images1[0].shape[0],
        )
        results = list()
        for image1, image2 in zip(images1, images2):

            # image diff
            image = image1.cuda() - image2.cuda()

            i = 255.0 * image.cpu().numpy()
            img = Image.fromarray(np.clip(i, 0, 255).astype(np.uint8))
            metadata = None
            file = f"{filename}_{counter:05}_.png"
            img.save(
                os.path.join(full_output_folder, file),
                pnginfo=metadata,
                compress_level=4,
            )
            results.append(
                {"filename": file, "subfolder": subfolder, "type": self.type}
            )
            counter += 1

            max_diff = image.abs().max().item()

            img1 = self.image_to_numpy(image1)
            img2 = self.image_to_numpy(image2)
            ssim = structural_similarity(img1, img2, channel_axis=2)

            print(
                "\033[91m"
                f"[ShowImageDiff {image_id}] Max value of diff is {max_diff}, image simularity is {ssim:.6f}"
                + "\033[0m"
            )

            if raise_if_diff == "enable":
                assert ssim > float(
                    ssim_threshold
                ), f"Image diff is too large, ssim is {ssim}, "

        return {"ui": {"images": results}}

    def image_to_numpy(self, image):
        i = 255.0 * image.cpu().numpy()
        img = Image.fromarray(np.clip(i, 0, 255).astype(np.uint8))
        return np.array(img)


NODE_CLASS_MAPPINGS = {
    "CompareModel": CompareModel,
    "ShowImageDiff": ShowImageDiff,
}

NODE_DISPLAY_NAME_MAPPINGS = {
    "CompareModel": "Model Weight Comparator",
    "ShowImageDiff": "Image Distinction Scanner",
}



MODEL = "runwayml/stable-diffusion-v1-5"
VARIANT = None
CUSTOM_PIPELINE = None
SCHEDULER = "EulerAncestralDiscreteScheduler"
LORA = None
CONTROLNET = None
STEPS = 30
PROMPT = "best quality, realistic, unreal engine, 4K, a beautiful girl"
NEGATIVE_PROMPT = ""
SEED = 333
WARMUPS = 1
BATCH = 1
HEIGHT = None
WIDTH = None
INPUT_IMAGE = None
CONTROL_IMAGE = None
OUTPUT_IMAGE = None
EXTRA_CALL_KWARGS = None
CACHE_INTERVAL = 3
CACHE_LAYER_ID = 0
CACHE_BLOCK_ID = 0
COMPILER = "oneflow"
COMPILER_CONFIG = None
QUANTIZE_CONFIG = None

import argparse
import importlib
import inspect
import json
import os
import time

import matplotlib.pyplot as plt
import numpy as np
import torch
from diffusers.utils import load_image
from onediff.infer_compiler import oneflow_compile

from onediffx import (  # quantize_pipe currently only supports the nexfort backend.
    compile_pipe,
    quantize_pipe,
)
from PIL import Image, ImageDraw


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("--model", type=str, default=MODEL)
    parser.add_argument("--variant", type=str, default=VARIANT)
    parser.add_argument("--custom-pipeline", type=str, default=CUSTOM_PIPELINE)
    parser.add_argument("--scheduler", type=str, default=SCHEDULER)
    parser.add_argument("--lora", type=str, default=LORA)
    parser.add_argument("--controlnet", type=str, default=CONTROLNET)
    parser.add_argument("--steps", type=int, default=STEPS)
    parser.add_argument("--prompt", type=str, default=PROMPT)
    parser.add_argument("--negative-prompt", type=str, default=NEGATIVE_PROMPT)
    parser.add_argument("--seed", type=int, default=SEED)
    parser.add_argument("--warmups", type=int, default=WARMUPS)
    parser.add_argument("--batch", type=int, default=BATCH)
    parser.add_argument("--height", type=int, default=HEIGHT)
    parser.add_argument("--width", type=int, default=WIDTH)
    parser.add_argument("--cache_interval", type=int, default=CACHE_INTERVAL)
    parser.add_argument("--cache_layer_id", type=int, default=CACHE_LAYER_ID)
    parser.add_argument("--cache_block_id", type=int, default=CACHE_BLOCK_ID)
    parser.add_argument("--extra-call-kwargs", type=str, default=EXTRA_CALL_KWARGS)
    parser.add_argument("--input-image", type=str, default=INPUT_IMAGE)
    parser.add_argument("--control-image", type=str, default=CONTROL_IMAGE)
    parser.add_argument("--output-image", type=str, default=OUTPUT_IMAGE)
    parser.add_argument("--print-output", action="store_true")
    parser.add_argument("--throughput", action="store_true")
    parser.add_argument("--deepcache", action="store_true")
    parser.add_argument(
        "--compiler",
        type=str,
        default=COMPILER,
        choices=["none", "oneflow", "nexfort", "compile", "compile-max-autotune"],
    )
    parser.add_argument(
        "--compiler-config",
        type=str,
        default=COMPILER_CONFIG,
    )
    parser.add_argument(
        "--run_multiple_resolutions",
        type=(lambda x: str(x).lower() in ["true", "1", "yes"]),
        default=False,
    )
    parser.add_argument("--quantize", action="store_true")
    parser.add_argument(
        "--quantize-config",
        type=str,
        default=QUANTIZE_CONFIG,
    )
    parser.add_argument("--quant-submodules-config-path", type=str, default=None)
    return parser.parse_args()


args = parse_args()


def load_pipe(
    pipeline_cls,
    model_name,
    variant=None,
    dtype=torch.float16,
    device="cuda",
    custom_pipeline=None,
    scheduler=None,
    lora=None,
    controlnet=None,
):
    extra_kwargs = {}
    if custom_pipeline is not None:
        extra_kwargs["custom_pipeline"] = custom_pipeline
    if variant is not None:
        extra_kwargs["variant"] = variant
    if dtype is not None:
        extra_kwargs["torch_dtype"] = dtype
    if controlnet is not None:
        from diffusers import ControlNetModel

        controlnet = ControlNetModel.from_pretrained(
            controlnet,
            torch_dtype=dtype,
        )
        extra_kwargs["controlnet"] = controlnet
    if os.path.exists(os.path.join(model_name, "calibrate_info.txt")):
        from onediff.quantization import QuantPipeline

        pipe = QuantPipeline.from_quantized(pipeline_cls, model_name, **extra_kwargs)
    else:
        pipe = pipeline_cls.from_pretrained(model_name, **extra_kwargs)
    if scheduler is not None and scheduler != "none":
        scheduler_cls = getattr(importlib.import_module("diffusers"), scheduler)
        pipe.scheduler = scheduler_cls.from_config(pipe.scheduler.config)
    if lora is not None:
        pipe.load_lora_weights(lora)
        pipe.fuse_lora()
    pipe.safety_checker = None
    if device is not None:
        pipe.to(torch.device(device))
    return pipe


class IterationProfiler:
    def __init__(self):
        self.begin = None
        self.end = None
        self.num_iterations = 0

    def get_iter_per_sec(self):
        if self.begin is None or self.end is None:
            return None
        self.end.synchronize()
        dur = self.begin.elapsed_time(self.end)
        return self.num_iterations / dur * 1000.0

    def callback_on_step_end(self, pipe, i, t, callback_kwargs={}):
        if self.begin is None:
            event = torch.cuda.Event(enable_timing=True)
            event.record()
            self.begin = event
        else:
            event = torch.cuda.Event(enable_timing=True)
            event.record()
            self.end = event
            self.num_iterations += 1
        return callback_kwargs


def calculate_inference_time_and_throughput(height, width, n_steps, model):
    start_time = time.time()
    model(prompt=args.prompt, height=height, width=width, num_inference_steps=n_steps)
    end_time = time.time()
    inference_time = end_time - start_time
    # pixels_processed = height * width * n_steps
    # throughput = pixels_processed / inference_time
    throughput = n_steps / inference_time
    return inference_time, throughput


def generate_data_and_fit_model(model, steps_range):
    height, width = 1024, 1024
    data = {"steps": [], "inference_time": [], "throughput": []}

    for n_steps in steps_range:
        inference_time, throughput = calculate_inference_time_and_throughput(
            height, width, n_steps, model
        )
        data["steps"].append(n_steps)
        data["inference_time"].append(inference_time)
        data["throughput"].append(throughput)
        print(
            f"Steps: {n_steps}, Inference Time: {inference_time:.2f} seconds, Throughput: {throughput:.2f} steps/s"
        )

    average_throughput = np.mean(data["throughput"])
    print(f"Average Throughput: {average_throughput:.2f} steps/s")

    coefficients = np.polyfit(data["steps"], data["inference_time"], 1)
    base_time_without_base_cost = 1 / coefficients[0]
    print(f"Throughput without base cost: {base_time_without_base_cost:.2f} steps/s")
    return data, coefficients


def plot_data_and_model(data, coefficients):
    plt.figure(figsize=(10, 5))
    plt.scatter(data["steps"], data["inference_time"], color="blue")
    plt.plot(data["steps"], np.polyval(coefficients, data["steps"]), color="red")
    plt.title("Inference Time vs. Steps")
    plt.xlabel("Steps")
    plt.ylabel("Inference Time (seconds)")
    plt.grid(True)
    # plt.savefig("output.png")
    plt.show()

    print(
        f"Model: Inference Time = {coefficients[0]:.2f} * Steps + {coefficients[1]:.2f}"
    )


def main():
    if args.input_image is None:
        if args.deepcache:
            from onediffx.deep_cache import StableDiffusionXLPipeline as pipeline_cls
        else:
            from diffusers import AutoPipelineForText2Image as pipeline_cls
    else:
        from diffusers import AutoPipelineForImage2Image as pipeline_cls

    pipe = load_pipe(
        pipeline_cls,
        args.model,
        variant=args.variant,
        custom_pipeline=args.custom_pipeline,
        scheduler=args.scheduler,
        lora=args.lora,
        controlnet=args.controlnet,
    )

    core_net = None
    if core_net is None:
        core_net = getattr(pipe, "unet", None)
    if core_net is None:
        core_net = getattr(pipe, "transformer", None)
    height = args.height or core_net.config.sample_size * pipe.vae_scale_factor
    width = args.width or core_net.config.sample_size * pipe.vae_scale_factor

    if args.compiler == "none":
        pass
    elif args.compiler == "oneflow":
        print("Oneflow backend is now active...")
        # Note: The compile_pipe() based on the oneflow backend is incompatible with T5EncoderModel.
        # pipe = compile_pipe(pipe)
        if hasattr(pipe, "unet"):
            pipe.unet = oneflow_compile(pipe.unet)
        if hasattr(pipe, "transformer"):
            pipe.transformer = oneflow_compile(pipe.transformer)
        pipe.vae.decoder = oneflow_compile(pipe.vae.decoder)
    elif args.compiler == "nexfort":
        print("Nexfort backend is now active...")
        if args.quantize:
            if args.quantize_config is not None:
                quantize_config = json.loads(args.quantize_config)
            else:
                quantize_config = '{"quant_type": "fp8_e4m3_e4m3_dynamic"}'
            if args.quant_submodules_config_path:
                # download: https://huggingface.co/siliconflow/PixArt-alpha-onediff-nexfort-fp8/blob/main/fp8_e4m3.json
                pipe = quantize_pipe(
                    pipe,
                    quant_submodules_config_path=args.quant_submodules_config_path,
                    ignores=[],
                    **quantize_config,
                )
            else:
                pipe = quantize_pipe(pipe, ignores=[], **quantize_config)
        if args.compiler_config is not None:
            # config with dict
            options = json.loads(args.compiler_config)
        else:
            # config with string
            options = '{"mode": "max-optimize:max-autotune:low-precision", "memory_format": "channels_last"}'
        pipe = compile_pipe(
            pipe, backend="nexfort", options=options, fuse_qkv_projections=True
        )
    elif args.compiler in ("compile", "compile-max-autotune"):
        mode = "max-autotune" if args.compiler == "compile-max-autotune" else None
        if hasattr(pipe, "unet"):
            pipe.unet = torch.compile(pipe.unet, mode=mode)
        if hasattr(pipe, "transformer"):
            pipe.transformer = torch.compile(pipe.transformer, mode=mode)
        if hasattr(pipe, "controlnet"):
            pipe.controlnet = torch.compile(pipe.controlnet, mode=mode)
        pipe.vae = torch.compile(pipe.vae, mode=mode)
    else:
        raise ValueError(f"Unknown compiler: {args.compiler}")

    if args.input_image is None:
        input_image = None
    else:
        input_image = load_image(args.input_image)
        input_image = input_image.resize((width, height), Image.LANCZOS)

    if args.control_image is None:
        if args.controlnet is None:
            control_image = None
        else:
            control_image = Image.new("RGB", (width, height))
            draw = ImageDraw.Draw(control_image)
            draw.ellipse(
                (args.width // 4, height // 4, args.width // 4 * 3, height // 4 * 3),
                fill=(255, 255, 255),
            )
            del draw
    else:
        control_image = load_image(args.control_image)
        control_image = control_image.resize((width, height), Image.LANCZOS)

    def get_kwarg_inputs():
        kwarg_inputs = dict(
            prompt=args.prompt,
            negative_prompt=args.negative_prompt,
            height=height,
            width=width,
            num_images_per_prompt=args.batch,
            generator=(
                None
                if args.seed is None
                else torch.Generator(device="cuda").manual_seed(args.seed)
            ),
            **(
                dict()
                if args.extra_call_kwargs is None
                else json.loads(args.extra_call_kwargs)
            ),
        )
        if args.steps is not None:
            kwarg_inputs["num_inference_steps"] = args.steps
        if input_image is not None:
            kwarg_inputs["image"] = input_image
        if control_image is not None:
            if input_image is None:
                kwarg_inputs["image"] = control_image
            else:
                kwarg_inputs["control_image"] = control_image
        if args.deepcache:
            kwarg_inputs["cache_interval"] = args.cache_interval
            kwarg_inputs["cache_layer_id"] = args.cache_layer_id
            kwarg_inputs["cache_block_id"] = args.cache_block_id
        return kwarg_inputs

    # NOTE: Warm it up.
    # The initial calls will trigger compilation and might be very slow.
    # After that, it should be very fast.
    if args.warmups > 0:
        begin = time.time()
        print("=======================================")
        print("Begin warmup")
        for _ in range(args.warmups):
            pipe(**get_kwarg_inputs())
        end = time.time()
        print("End warmup")
        print(f"Warmup time: {end - begin:.3f}s")
        print("=======================================")

    # Let"s see it!
    # Note: Progress bar might work incorrectly due to the async nature of CUDA.
    kwarg_inputs = get_kwarg_inputs()
    iter_profiler = IterationProfiler()
    if "callback_on_step_end" in inspect.signature(pipe).parameters:
        kwarg_inputs["callback_on_step_end"] = iter_profiler.callback_on_step_end
    elif "callback" in inspect.signature(pipe).parameters:
        kwarg_inputs["callback"] = iter_profiler.callback_on_step_end
    begin = time.time()
    output_images = pipe(**kwarg_inputs).images
    end = time.time()

    print("=======================================")
    print(f"Inference time: {end - begin:.3f}s")
    iter_per_sec = iter_profiler.get_iter_per_sec()
    if iter_per_sec is not None:
        print(f"Iterations per second: {iter_per_sec:.3f}")
    if args.compiler == "oneflow":
        import oneflow as flow  # usort: skip

        cuda_mem_after_used = flow._oneflow_internal.GetCUDAMemoryUsed() / 1024
    else:
        cuda_mem_after_used = torch.cuda.max_memory_allocated() / (1024**3)
    print(f"Max used CUDA memory : {cuda_mem_after_used:.3f}GiB")
    print("=======================================")

    if args.print_output:
        from onediff.utils.import_utils import is_nexfort_available

        if is_nexfort_available():
            from nexfort.utils.term_image import print_image

            for image in output_images:
                print_image(image, max_width=80)

    if args.output_image is not None:
        output_images[0].save(args.output_image)
    else:
        print("Please set `--output-image` to save the output image")

    if args.run_multiple_resolutions:
        print("Test run with multiple resolutions...")
        sizes = [1024, 512, 768, 256]
        for h in sizes:
            for w in sizes:
                kwarg_inputs["height"] = h
                kwarg_inputs["width"] = w
                print(f"Running at resolution: {h}x{w}")
                start_time = time.time()
                image = pipe(**kwarg_inputs).images
                end_time = time.time()
                print(f"Inference time: {end_time - start_time:.2f} seconds")

    if args.throughput:
        steps_range = range(1, 100, 1)
        data, coefficients = generate_data_and_fit_model(pipe, steps_range)
        plot_data_and_model(data, coefficients)


if __name__ == "__main__":
    main()



from typing import Dict


def process_state_dict_before_saving(state_dict: Dict, graph=None):
    assert graph is not None

    try:
        for graph_key, oneflow_graph in state_dict.items():
            attn2_patches = (
                oneflow_graph.get("inputs_original", [])[1]
                .get("transformer_options", {})
                .get("patches_replace", {})
                .get("attn2")
            )
            if attn2_patches:

                del oneflow_graph["inputs_original"][1]["transformer_options"][
                    "patches_replace"
                ]["attn2"]

    except Exception as e:
        print(f"Warning: Failed to process state dict before saving: {e}")

    return state_dict


def attn2_patch_sum(input_kwargs) -> int:
    attn2 = (
        input_kwargs.get("transformer_options", {})
        .get("patches_replace", {})
        .get("attn2", {})
    )
    return len(attn2)



import hashlib
import os
from pathlib import Path

# ComfyUI
from folder_paths import get_input_directory

# onediff
from onediff import __version__ as onediff_version
from oneflow import __version__ as oneflow_version


def generate_short_sha256(string: str) -> str:
    return hashlib.sha256(string.encode("utf-8")).hexdigest()[:10]


def generate_graph_path(ckpt_name, model) -> Path:
    # get_save_graph_directory
    default_dir = get_input_directory()
    input_dir = os.getenv("COMFYUI_ONEDIFF_SAVE_GRAPH_DIR", default_dir)

    input_dir = Path(input_dir)
    graph_dir = input_dir / "graphs"

    file_name = f"{type(model).__name__}"

    return graph_dir / file_name



from typing import Any, Mapping

import torch
from modules import sd_models
from modules.sd_hijack_utils import CondFunc
from onediff_shared import onediff_enabled


def unload_model_weights(sd_model=None, info=None):
    from modules import devices, lowvram, shared

    m = sd_model or shared.sd_model
    if m.lowvram:
        lowvram.send_everything_to_cpu()
    else:
        m.to(devices.cpu)
    devices.torch_gc()
    return sd_model


def send_model_to_cpu(m):
    # do nothing
    pass


def hijack_function(module, name, new_name, new_value):
    # restore original function in case of reload
    unhijack_function(module=module, name=name, new_name=new_name)
    setattr(module, new_name, getattr(module, name))
    setattr(module, name, new_value)


def unhijack_function(module, name, new_name):
    if hasattr(module, new_name):
        setattr(module, name, getattr(module, new_name))
        delattr(module, new_name)


def do_hijack():
    from modules import script_callbacks, sd_models

    script_callbacks.on_script_unloaded(undo_hijack)
    hijack_function(
        module=sd_models,
        name="unload_model_weights",
        new_name="__onediff_original_unload_model_weights",
        new_value=unload_model_weights,
    )
    hijack_function(
        module=sd_models,
        name="send_model_to_cpu",
        new_name="__onediff_original_send_model_to_cpu",
        new_value=send_model_to_cpu,
    )


def undo_hijack():
    from modules import sd_models

    unhijack_function(
        module=sd_models,
        name="unload_model_weights",
        new_name="__onediff_original_unload_model_weights",
    )
    unhijack_function(
        module=sd_models,
        name="send_model_to_cpu",
        new_name="__onediff_original_send_model_to_cpu",
    )


def onediff_hijack_load_model_weights(
    orig_func, model, checkpoint_info: sd_models.CheckpointInfo, state_dict: dict, timer
):
    # load_model_weights(model, checkpoint_info: CheckpointInfo, state_dict, timer)
    sd_model_hash = checkpoint_info.calculate_shorthash()
    import onediff_shared

    if onediff_shared.current_unet_graph.sha == sd_model_hash:
        model.model.diffusion_model = onediff_shared.current_unet_graph.graph_module
        state_dict = {
            k: v
            for k, v in state_dict.items()
            if not k.startswith("model.diffusion_model.")
        }

        # for stable-diffusion-webui/modules/sd_models.py:load_model_weights model.is_ssd check
        state_dict[
            "model.diffusion_model.middle_block.1.transformer_blocks.0.attn1.to_q.weight"
        ] = model.get_parameter(
            "model.diffusion_model.middle_block.1.transformer_blocks.0.attn1.to_q.weight"
        )
    return orig_func(model, checkpoint_info, state_dict, timer)


def onediff_hijack_load_state_dict(
    orig_func,
    self,
    state_dict: Mapping[str, Any],
    strict: bool = True,
    assign: bool = False,
):
    if (
        len(state_dict) > 0
        and next(iter(state_dict.values())).is_cuda
        and next(self.parameters()).is_meta
    ):
        return orig_func(self, state_dict, strict, assign=True)
    else:
        return orig_func(self, state_dict, strict, assign)


# fmt: off
def onediff_hijaced_LoadStateDictOnMeta___enter__(orig_func, self):
    from modules import shared
    if shared.cmd_opts.disable_model_loading_ram_optimization:
        return

    sd = self.state_dict
    device = self.device

    def load_from_state_dict(original, module, state_dict, prefix, *args, **kwargs):
        used_param_keys = []

        for name, param in module._parameters.items():
            if param is None:
                continue

            key = prefix + name
            sd_param = sd.pop(key, None)
            if sd_param is not None:
                state_dict[key] = sd_param.to(dtype=self.get_weight_dtype(key))
                used_param_keys.append(key)

            if param.is_meta:
                dtype = sd_param.dtype if sd_param is not None else param.dtype
                module._parameters[name] = torch.nn.parameter.Parameter(torch.zeros_like(param, device=device, dtype=dtype), requires_grad=param.requires_grad)

        for name in module._buffers:
            key = prefix + name

            sd_param = sd.pop(key, None)
            if sd_param is not None:
                state_dict[key] = sd_param
                used_param_keys.append(key)

        original(module, state_dict, prefix, *args, **kwargs)

        for key in used_param_keys:
            state_dict.pop(key, None)

    # def load_state_dict(original, module, state_dict, strict=True):
    def load_state_dict(original, module, state_dict, strict=True):
        """torch makes a lot of copies of the dictionary with weights, so just deleting entries from state_dict does not help
        because the same values are stored in multiple copies of the dict. The trick used here is to give torch a dict with
        all weights on meta device, i.e. deleted, and then it doesn't matter how many copies torch makes.

        In _load_from_state_dict, the correct weight will be obtained from a single dict with the right weights (sd).

        The dangerous thing about this is if _load_from_state_dict is not called, (if some exotic module overloads
        the function and does not call the original) the state dict will just fail to load because weights
        would be on the meta device.
        """

        if state_dict is sd:
            state_dict = {k: v.to(device="meta", dtype=v.dtype) for k, v in state_dict.items()}

        # ------------------- DIFF HERE -------------------
        # original(module, state_dict, strict=strict)
        if len(state_dict) > 0 and next(iter(state_dict.values())).is_cuda and next(module.parameters()).is_meta:
            assign = True
        else:
            assign = False
        # orig_func(original, module, state_dict, strict=strict, assign=assign)
        original(module, state_dict, strict=strict, assign=assign)

    module_load_state_dict = self.replace(torch.nn.Module, 'load_state_dict', lambda *args, **kwargs: load_state_dict(module_load_state_dict, *args, **kwargs))
    module_load_from_state_dict = self.replace(torch.nn.Module, '_load_from_state_dict', lambda *args, **kwargs: load_from_state_dict(module_load_from_state_dict, *args, **kwargs))
    linear_load_from_state_dict = self.replace(torch.nn.Linear, '_load_from_state_dict', lambda *args, **kwargs: load_from_state_dict(linear_load_from_state_dict, *args, **kwargs))
    conv2d_load_from_state_dict = self.replace(torch.nn.Conv2d, '_load_from_state_dict', lambda *args, **kwargs: load_from_state_dict(conv2d_load_from_state_dict, *args, **kwargs))
    mha_load_from_state_dict = self.replace(torch.nn.MultiheadAttention, '_load_from_state_dict', lambda *args, **kwargs: load_from_state_dict(mha_load_from_state_dict, *args, **kwargs))
    layer_norm_load_from_state_dict = self.replace(torch.nn.LayerNorm, '_load_from_state_dict', lambda *args, **kwargs: load_from_state_dict(layer_norm_load_from_state_dict, *args, **kwargs))
    group_norm_load_from_state_dict = self.replace(torch.nn.GroupNorm, '_load_from_state_dict', lambda *args, **kwargs: load_from_state_dict(group_norm_load_from_state_dict, *args, **kwargs))
# fmt: on


CondFunc(
    "modules.sd_disable_initialization.LoadStateDictOnMeta.__enter__",
    onediff_hijaced_LoadStateDictOnMeta___enter__,
    lambda _, *args, **kwargs: onediff_enabled,
)
CondFunc(
    "modules.sd_models.load_model_weights",
    onediff_hijack_load_model_weights,
    lambda _, *args, **kwargs: onediff_enabled,
)



from .booster_basic import BasicOneFlowBoosterExecutor
from .booster_deepcache import DeepcacheBoosterExecutor
from .booster_patch import PatchBoosterExecutor
from .config import _USE_UNET_INT8, ONEDIFF_QUANTIZED_OPTIMIZED_MODELS
from .patch_management.patch_for_oneflow import *

from .hijack_animatediff import animatediff_hijacker
from .hijack_comfyui_instantid import comfyui_instantid_hijacker
from .hijack_ipadapter_plus import ipadapter_plus_hijacker
from .hijack_model_management import model_management_hijacker
from .hijack_model_patcher import model_patch_hijacker
from .hijack_nodes import nodes_hijacker
from .hijack_pulid_comfyui import pulid_comfyui_hijacker
from .hijack_samplers import samplers_hijack
from .hijack_utils import comfy_utils_hijack

model_management_hijacker.hijack()  # add flow.cuda.empty_cache()
nodes_hijacker.hijack()
samplers_hijack.hijack()
animatediff_hijacker.hijack()
ipadapter_plus_hijacker.hijack()
comfyui_instantid_hijacker.hijack()
model_patch_hijacker.hijack()
comfy_utils_hijack.hijack()
pulid_comfyui_hijacker.hijack()



import collections
import os

import yaml
from onediff.infer_compiler.backends.nexfort import fallback_to_eager

from ..modules.nexfort.booster_basic import BasicNexFortBoosterExecutor


# https://github.com/siliconflow/nexfort?tab=readme-ov-file#suggested-combinations-of-compiler-modes
compiler_modes = collections.OrderedDict(
    {
        "max-optimize:max-autotune:low-precision": "This will deliver a good performance and adapt quickly to shape changes.",
        "max-optimize:max-autotune:low-precision:freezing:benchmark": "",
        "jit:disable-runtime-fusion:low-precision": "This compiles super quickly, but the performance might not be optimized very noticeably.",
        "jit:benchmark:low-precision:freezing:cudagraphs": "This compiles the model very quickly, but the performance might be not as good as `TorchInductor` optimized models.",
        "max-autotune:benchmark:low-precision:cudagraphs": "This is the most suggested combination of compiler modes. It will deliver a good balance between performance and compilation time.",
        "max-optimize:max-autotune:benchmark:low-precision:freezing:cudagraphs": "This is the most aggressive combination of compiler modes. It will deliver the best performance but might slow down the compilation significantly.",
    }
)


class OneDiffNexfortBooster:
    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "fullgraph": ([False, True],),
                "dynamic": ([True, False],),
                "mode": ([mode for mode in compiler_modes.keys()],),
                "docs_link": (
                    "STRING",
                    {
                        "multiline": True,
                        "default": "[Note]: \nInstall-nexfort \nhttps://github.com/siliconflow/onediff/tree/main/src/onediff/infer_compiler/backends/nexfort#install-nexfort",
                    },
                ),
            }
        }

    CATEGORY = "OneDiff/Booster"
    RETURN_TYPES = ("TorchCompileBooster",)
    FUNCTION = "apply"

    def apply(
        self,
        fullgraph=False,
        dynamic=None,
        mode="max-autotune:cudagraphs",
        docs_link=None,
    ):
        return (
            BasicNexFortBoosterExecutor(
                fullgraph=fullgraph, mode=f"{mode}:cache-all", dynamic=dynamic
            ),
        )


class ResolutionSpeedupChecker:
    current_dir = os.path.dirname(os.path.abspath(__file__))
    config_path = os.path.join(current_dir, "resolutions_config.yaml")
    with open(config_path, "r") as file:
        resolutions = yaml.safe_load(file)
    height_width_dict = {x["height"]: x["width"] for x in resolutions["resolutions"]}

    @classmethod
    def INPUT_TYPES(s):
        return {
            "required": {
                "samples": ("LATENT",),
            }
        }

    RETURN_TYPES = ("LATENT",)
    FUNCTION = "check"

    def check(self, samples):
        _, _, H, W = samples["samples"].shape
        H, W = H * 8, W * 8
        if H in self.height_width_dict and self.height_width_dict[H] == W:
            fallback_to_eager(True)
        else:
            fallback_to_eager(False)
        return (samples,)


NODE_CLASS_MAPPINGS = {
    "OneDiffNexfortBooster": OneDiffNexfortBooster,
    "ResolutionSpeedupChecker": ResolutionSpeedupChecker,
}

NODE_DISPLAY_NAME_MAPPINGS = {
    "OneDiffNexfortBooster": "Nexfort Booster - OneDiff",
    "ResolutionSpeedupChecker": "Speedup Checker - Resolution",
}



import argparse
import os

from PIL import Image
from PIL.PngImagePlugin import PngInfo

usage = """
    python meta2png.py --key api --png image.png --api api.json --output output.png OR
    python meta2png.py --key api --png image.png
"""


def parse_args():
    parser = argparse.ArgumentParser(description=usage)
    parser.add_argument(
        "--api",
        type=str,
        default="",
        help="The json file containing the workflow for API calling.",
    )
    parser.add_argument(
        "--png",
        type=str,
        help="The png file containing the UI workflow meta data.",
        required=True,
    )
    parser.add_argument(
        "--output",
        type=str,
        default="",
        help="The output png filename.",
    )
    parser.add_argument(
        "--key",
        type=str,
        default="api",
        help="The key name of the metadata. Default: api",
    )

    args = parser.parse_args()
    return args


def add_meta2png(args):
    origin_img = Image.open(args.png)
    info_dict = origin_img.info
    metadata = PngInfo()
    if info_dict is not None:
        for k, v in info_dict.items():
            metadata.add_text(k, str(v))

    with open(args.api, "r") as f:
        c = f.read()
        metadata.add_text(args.key, c)

    origin_img.close()

    api_img = Image.open(args.png)

    if args.output == "":
        pngname, _ = os.path.splitext(os.path.basename(args.png))
        apiname, _ = os.path.splitext(os.path.basename(args.api))
        output = f"processed-{apiname}-{pngname}.png"
    else:
        output = args.output
    api_img.save(output, pnginfo=metadata)
    api_img.close()


def show_meta4png(args):
    origin_img = Image.open(args.png)
    info_dict = origin_img.info

    if info_dict is None:
        print(f"There is NO meta data in {args.png} found")
        return

    if not args.key in info_dict.keys():
        print(f"{args.key} is not found in {info_dict.keys()}")
        return

    print(info_dict[args.key])


if __name__ == "__main__":
    args = parse_args()
    if args.api != "":  # no josn file specified
        add_meta2png(args)
    else:
        show_meta4png(args)



import os
from contextlib import contextmanager
from pathlib import Path
from textwrap import dedent
from zipfile import BadZipFile

import onediff_shared

from onediff.infer_compiler import DeployableModule

hints_message = dedent(
    """\
<div id="hintMessage" style="position: relative; padding: 20px; border: 1px solid #e0e0e0; border-radius: 5px; background-color: #f9f9f9;">
    <button onclick="document.getElementById('hintMessage').style.display = 'none'" style="position: absolute; top: 10px; right: 10px; background: none; border: none; font-size: 18px; cursor: pointer;">&times;</button>
    <div style="font-size: 18px; font-weight: bold; margin-bottom: 15px; color: #31708f;">
        Hints Message
    </div>
    <div style="padding: 10px; border: 1px solid #31708f; border-radius: 5px; background-color: #f9f9f9;">
        Hints: Enterprise function is not supported on your system.
    </div>
    <p style="margin-top: 15px;">
        If you need Enterprise Level Support for your system or business, please send an email to
        <a href="mailto:business@siliconflow.com" style="color: #31708f; text-decoration: none;">business@siliconflow.com</a>.
        <br>
        Tell us about your use case, deployment scale, and requirements.
    </p>
    <p>
        <strong>GitHub Issue:</strong>
        <a href="https://github.com/siliconflow/onediff/issues" style="color: #31708f; text-decoration: none;">https://github.com/siliconflow/onediff/issues</a>
    </p>
</div>
"""
)

all_compiler_caches = []


def all_compiler_caches_path():
    import modules.shared as shared

    caches_path = Path(shared.opts.onediff_compiler_caches_path)
    if not caches_path.exists():
        caches_path.mkdir(parents=True)
    return shared.opts.onediff_compiler_caches_path


def get_all_compiler_caches():
    global all_compiler_caches
    if len(all_compiler_caches) == 0:
        refresh_all_compiler_caches()
    return all_compiler_caches


def refresh_all_compiler_caches(path: Path = None):
    global all_compiler_caches
    path = path or all_compiler_caches_path()
    all_compiler_caches = [f.stem for f in Path(path).iterdir() if f.is_file()]


def load_graph(compiled_unet: DeployableModule, compiler_cache: str):
    from compile import OneDiffCompiledGraph

    if isinstance(compiled_unet, OneDiffCompiledGraph):
        compiled_unet = compiled_unet.graph_module

    if compiler_cache in [None, "None"]:
        return

    compiler_cache_path = all_compiler_caches_path() + f"/{compiler_cache}"
    if not Path(compiler_cache_path).exists():
        raise FileNotFoundError(
            f"Cannot find cache {compiler_cache_path}, please make sure it exists"
        )
    try:
        compiled_unet.load_graph(compiler_cache_path, run_warmup=True)
    except BadZipFile:
        raise RuntimeError(
            "Load cache failed. Please make sure that the --disable-safe-unpickle parameter is added when starting the webui"
        )
    except Exception as e:
        raise RuntimeError(
            f"Load cache failed ({e}). Please make sure cache has the same sd version (or unet architure) with current checkpoint"
        )
    return compiled_unet


def save_graph(compiled_unet: DeployableModule, saved_cache_name: str = ""):
    from compile import OneDiffCompiledGraph

    if isinstance(compiled_unet, OneDiffCompiledGraph):
        compiled_unet = compiled_unet.graph_module

    if saved_cache_name in ["", None]:
        return

    if not os.access(str(all_compiler_caches_path()), os.W_OK):
        raise PermissionError(
            f"The directory {all_compiler_caches_path()} does not have write permissions, and compiler cache cannot be written to this directory. \
                                Please change it in the settings to a directory with write permissions"
        )
    if not Path(all_compiler_caches_path()).exists():
        Path(all_compiler_caches_path()).mkdir()
    saved_cache_name = all_compiler_caches_path() + f"/{saved_cache_name}"
    if not Path(saved_cache_name).exists():
        compiled_unet.save_graph(saved_cache_name)


@contextmanager
def onediff_enabled():
    onediff_shared.onediff_enabled = True
    try:
        yield
    finally:
        onediff_shared.onediff_enabled = False



try:
    from ._version import version as __version__, version_tuple
except ImportError:
    __version__ = "unknown version"
    version_tuple = (0, 0, "unknown version")

__author__ = "OneDiff"
__credits__ = "OneDiff contributors"



def check_if_controlnet_ext_loaded() -> bool:
    from modules import extensions

    return "sd-webui-controlnet" in extensions.loaded_extensions


def get_controlnet_script(p):
    for script in p.scripts.scripts:
        if script.__module__ == "controlnet.py":
            return script
    return None


def check_if_controlnet_enabled(p):
    controlnet_script_class = get_controlnet_script(p)
    return (
        controlnet_script_class is not None
        and len(controlnet_script_class.get_enabled_units(p)) != 0
    )



import os

from .hijack_comfyui_instantid import comfyui_instantid_hijacker
from .hijack_ipadapter_plus import ipadapter_plus_hijacker
from .hijack_pulid_comfyui import pulid_comfyui_hijacker
from .hijack_samplers import samplers_hijack

samplers_hijack.hijack(last=False)
ipadapter_plus_hijacker.hijack(last=False)
pulid_comfyui_hijacker.hijack(last=False)
comfyui_instantid_hijacker.hijack(last=False)


# https://github.com/pytorch/pytorch/blob/1edcb31d34ef012d828bb9f39a8aef6020f580b2/aten/src/ATen/cuda/CUDABlas.cpp#L182-L203
if os.getenv("CUBLASLT_WORKSPACE_SIZE") is None:
    os.environ["CUBLASLT_WORKSPACE_SIZE"] = str(1024 * 1024)



import argparse

import cv2
import numpy as np
import torch

from diffusers import (
    ControlNetModel,
    StableDiffusionControlNetPipeline,
    UniPCMultistepScheduler,
)
from diffusers.utils import load_image
from PIL import Image

parser = argparse.ArgumentParser()
parser.add_argument("--base", type=str, default="runwayml/stable-diffusion-v1-5")
parser.add_argument("--controlnet", type=str, default="lllyasviel/sd-controlnet-canny")
parser.add_argument(
    "--input_image",
    type=str,
    default="https://hf.co/datasets/huggingface/documentation-images/resolve/main/diffusers/input_image_vermeer.png",
)
parser.add_argument(
    "--prompt",
    type=str,
    default="chinese painting style women",
)
parser.add_argument("--height", type=int, default=512)
parser.add_argument("--width", type=int, default=512)
parser.add_argument("--n_steps", type=int, default=30)
parser.add_argument(
    "--saved_image", type=str, required=False, default="controlnet-out.png"
)
parser.add_argument("--seed", type=int, default=1)
parser.add_argument("--warmup", type=int, default=1)
parser.add_argument("--run", type=int, default=3)
parser.add_argument(
    "--compile", type=(lambda x: str(x).lower() in ["true", "1", "yes"]), default=True
)
parser.add_argument(
    "--compile_ctrlnet",
    type=(lambda x: str(x).lower() in ["true", "1", "yes"]),
    default=True,
)
args = parser.parse_args()

# load an image
image = load_image(args.input_image)
image = np.array(image)

# get canny image
image = cv2.Canny(image, 100, 200)
image = image[:, :, None]
image = np.concatenate([image, image, image], axis=2)
canny_image = Image.fromarray(image)

# load control net and stable diffusion
# reference: https://huggingface.co/docs/diffusers/main/en/api/pipelines/controlnet
controlnet = ControlNetModel.from_pretrained(args.controlnet, torch_dtype=torch.float16)
pipe = StableDiffusionControlNetPipeline.from_pretrained(
    args.base, controlnet=controlnet, torch_dtype=torch.float16
)

# speed up diffusion process with faster scheduler and memory optimization
pipe.scheduler = UniPCMultistepScheduler.from_config(pipe.scheduler.config)
pipe.to("cuda")

# remove following line if xformers is not installed
# pipe.enable_xformers_memory_efficient_attention()

if args.compile:
    from onediff.infer_compiler import oneflow_compile

    pipe.unet = oneflow_compile(pipe.unet)
    pipe.vae.decoder = oneflow_compile(pipe.vae.decoder)
if args.compile_ctrlnet:
    from onediff.infer_compiler import oneflow_compile

    pipe.controlnet = oneflow_compile(pipe.controlnet)


# generate image
generator = torch.manual_seed(args.seed)

print("Warmup")
for i in range(args.warmup):
    images = pipe(
        args.prompt,
        height=args.height,
        width=args.width,
        num_inference_steps=args.n_steps,
        generator=generator,
        image=canny_image,
    ).images

print("Run")
for i in range(args.run):
    image = pipe(
        args.prompt,
        height=args.height,
        width=args.width,
        num_inference_steps=args.n_steps,
        generator=generator,
        image=canny_image,
    ).images[0]
    image.save(f"{i=}th_{args.saved_image}.png")



import argparse
import json
import time

import torch
from diffusers import StableDiffusion3Pipeline
from onediffx import compile_pipe, quantize_pipe


def parse_args():
    parser = argparse.ArgumentParser(
        description="Use onediif (nexfort) to accelerate image generation with Stable Diffusion 3."
    )
    parser.add_argument(
        "--model",
        type=str,
        default="stabilityai/stable-diffusion-3-medium-diffusers",
        help="Model path or identifier.",
    )
    parser.add_argument(
        "--compiler-config", type=str, help="JSON string for compiler config."
    )
    parser.add_argument(
        "--quantize-config", type=str, help="JSON string for quantization config."
    )
    parser.add_argument(
        "--prompt",
        type=str,
        default="photo of a dog and a cat both standing on a red box, with a blue ball in the middle with a parrot standing on top of the ball. The box has the text 'onediff'",
        help="Prompt for the image generation.",
    )
    parser.add_argument(
        "--negative_prompt",
        type=str,
        default="",
        help="Negative prompt for the image generation.",
    )
    parser.add_argument(
        "--height", type=int, default=1024, help="Height of the generated image."
    )
    parser.add_argument(
        "--width", type=int, default=1024, help="Width of the generated image."
    )
    parser.add_argument(
        "--guidance_scale",
        type=float,
        default=4.5,
        help="The scale factor for the guidance.",
    )
    parser.add_argument(
        "--num-inference-steps", type=int, default=28, help="Number of inference steps."
    )
    parser.add_argument(
        "--saved-image",
        type=str,
        default="./sd3.png",
        help="Path to save the generated image.",
    )
    parser.add_argument(
        "--seed", type=int, default=1, help="Seed for random number generation."
    )
    parser.add_argument(
        "--run_multiple_resolutions",
        type=(lambda x: str(x).lower() in ["true", "1", "yes"]),
        default=False,
    )
    parser.add_argument(
        "--run_multiple_prompts",
        type=(lambda x: str(x).lower() in ["true", "1", "yes"]),
        default=False,
    )
    return parser.parse_args()


args = parse_args()

device = torch.device("cuda")


def generate_texts(min_length=50, max_length=302):
    base_text = "a female character with long, flowing hair that appears to be made of ethereal, swirling patterns resembling the Northern Lights or Aurora Borealis. The background is dominated by deep blues and purples, creating a mysterious and dramatic atmosphere. The character's face is serene, with pale skin and striking features. She"

    additional_words = [
        "gracefully",
        "beautifully",
        "elegant",
        "radiant",
        "mysteriously",
        "vibrant",
        "softly",
        "gently",
        "luminescent",
        "sparkling",
        "delicately",
        "glowing",
        "brightly",
        "shimmering",
        "enchanting",
        "gloriously",
        "magnificent",
        "majestic",
        "fantastically",
        "dazzlingly",
    ]

    for i in range(min_length, max_length):
        idx = i % len(additional_words)
        base_text += " " + additional_words[idx]
        yield base_text


class SD3Generator:
    def __init__(self, model, compiler_config=None, quantize_config=None):
        self.pipe = StableDiffusion3Pipeline.from_pretrained(
            model,
            torch_dtype=torch.float16,
        )
        self.pipe.to(device)

        if compiler_config:
            print("compile...")
            self.pipe = self.compile_pipe(self.pipe, compiler_config)

        if quantize_config:
            print("quant...")
            self.pipe = self.quantize_pipe(self.pipe, quantize_config)

    def warmup(self, gen_args, warmup_iterations=1):
        warmup_args = gen_args.copy()

        warmup_args["generator"] = torch.Generator(device=device).manual_seed(0)

        print("Starting warmup...")
        start_time = time.time()
        for _ in range(warmup_iterations):
            self.pipe(**warmup_args)
        end_time = time.time()
        print("Warmup complete.")
        print(f"Warmup time: {end_time - start_time:.2f} seconds")

    def generate(self, gen_args):
        gen_args["generator"] = torch.Generator(device=device).manual_seed(args.seed)

        # Run the model
        start_time = time.time()
        images = self.pipe(**gen_args).images
        end_time = time.time()

        images[0].save(args.saved_image)

        return images[0], end_time - start_time

    def compile_pipe(self, pipe, compiler_config):
        options = compiler_config
        pipe = compile_pipe(
            pipe, backend="nexfort", options=options, fuse_qkv_projections=True
        )
        return pipe

    def quantize_pipe(self, pipe, quantize_config):
        pipe = quantize_pipe(pipe, ignores=[], **quantize_config)
        return pipe


def main():
    compiler_config = json.loads(args.compiler_config) if args.compiler_config else None
    quantize_config = json.loads(args.quantize_config) if args.quantize_config else None

    sd3 = SD3Generator(args.model, compiler_config, quantize_config)

    if args.run_multiple_prompts:
        # Note: diffusers will truncate the input prompt (limited to 77 tokens).
        # https://github.com/huggingface/diffusers/blob/8e1b7a084addc4711b8d9be2738441dfad680ce0/src/diffusers/pipelines/stable_diffusion_3/pipeline_stable_diffusion_3.py#L238
        dynamic_prompts = generate_texts(max_length=101)
        prompt_list = list(dynamic_prompts)
    else:
        prompt_list = [args.prompt]

    gen_args = {
        "prompt": args.prompt,
        "num_inference_steps": args.num_inference_steps,
        "height": args.height,
        "width": args.width,
        "guidance_scale": args.guidance_scale,
        "negative_prompt": args.negative_prompt,
    }

    sd3.warmup(gen_args)

    for prompt in prompt_list:
        gen_args["prompt"] = prompt
        print(f"Processing prompt of length {len(prompt)} characters.")
        image, inference_time = sd3.generate(gen_args)
        assert inference_time < 20, "Prompt inference took too long"
        print(
            f"Generated image saved to {args.saved_image} in {inference_time:.2f} seconds."
        )
        cuda_mem_after_used = torch.cuda.max_memory_allocated() / (1024**3)
        print(f"Max used CUDA memory : {cuda_mem_after_used:.3f} GiB")

    if args.run_multiple_resolutions:
        gen_args["prompt"] = args.prompt
        print("Test run with multiple resolutions...")
        sizes = [1536, 1024, 768, 720, 576, 512, 256]
        for h in sizes:
            for w in sizes:
                gen_args["height"] = h
                gen_args["width"] = w
                print(f"Running at resolution: {h}x{w}")
                start_time = time.time()
                sd3.generate(gen_args)
                end_time = time.time()
                print(f"Inference time: {end_time - start_time:.2f} seconds")
                assert (
                    end_time - start_time
                ) < 20, "Resolution switch test took too long"


if __name__ == "__main__":
    main()



import time
from pathlib import Path

import pandas as pd
import safetensors.torch

import torch
from diffusers import DiffusionPipeline

from onediff.infer_compiler import oneflow_compile
from onediff.torch_utils import TensorInplaceAssign
from onediffx.lora import load_and_fuse_lora, unfuse_lora

_time = None


class TimerContextManager:
    def __init__(self, msg, lora):
        self.msg = msg
        self.lora_name = lora

    def __enter__(self):
        self.start_time = time.perf_counter()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.end_time = time.perf_counter()
        elapsed_time = self.end_time - self.start_time
        global _time
        _time = elapsed_time
        print(
            f"Time cost {elapsed_time:.2f}, of method {self.msg}, lora name {self.lora_name}"
        )


MODEL_ID = "/share_nfs/hf_models/stable-diffusion-xl-base-1.0"
pipe = DiffusionPipeline.from_pretrained(
    MODEL_ID, variant="fp16", torch_dtype=torch.float16
).to("cuda")

LORA_MODEL_ID = [
    "/share_nfs/onediff_ci/diffusers/loras/SDXL-Emoji-Lora-r4.safetensors",
    "/share_nfs/onediff_ci/diffusers/loras/sdxl_metal_lora.safetensors",
    "/share_nfs/onediff_ci/diffusers/loras/simple_drawing_xl_b1-000012.safetensors",
    "/share_nfs/onediff_ci/diffusers/loras/texta.safetensors",
    "/share_nfs/onediff_ci/diffusers/loras/watercolor_v1_sdxl_lora.safetensors",
]

loras = {name: safetensors.torch.load_file(name) for name in LORA_MODEL_ID}

pipe.unet = oneflow_compile(pipe.unet)
generator = torch.manual_seed(0)

load_lora_weights_time = []
for i, (name, lora) in enumerate(loras.items()):
    with TimerContextManager("load_lora_weights", Path(name).stem):
        pipe.load_lora_weights(lora.copy())
    load_lora_weights_time.append(_time)

modules = [pipe.unet, pipe.text_encoder]
if hasattr(pipe, "text_encoder_2"):
    modules.append(pipe.text_encoder_2)

print("")
load_lora_weights_and_fuse_lora_time = []
for i, (name, lora) in enumerate(loras.items()):
    with TimerContextManager("load_lora_weights and fuse_lora", Path(name).stem):
        pipe.load_lora_weights(lora.copy())
        with TensorInplaceAssign(*modules):
            pipe.fuse_lora(lora_scale=1.0)
            pipe.unfuse_lora()
    load_lora_weights_and_fuse_lora_time.append(_time)

print("")
load_and_fuse_lora_time = []
for i, (name, lora) in enumerate(loras.items()):
    with TimerContextManager("load_and_fuse_lora", Path(name).stem):
        load_and_fuse_lora(
            pipe, lora.copy(), adapter_name=Path(name).stem, lora_scale=1.0
        )
        unfuse_lora(pipe)
    load_and_fuse_lora_time.append(_time)

data = {
    "LoRA name": [Path(x).stem for x in loras],
    "size": ["28M", "23M", "55M", "270M", "12M"],
    "load_lora_weights": [f"{x:.2f} s" for x in load_lora_weights_time],
    "load_lora_weights + fuse_lora": [
        f"{x:.2f} s" for x in load_lora_weights_and_fuse_lora_time
    ],
    "onediffx load_and_fuse_lora": [f"{x:.2f} s" for x in load_and_fuse_lora_time],
    "src link": [
        "[Link](https://novita.ai/model/SDXL-Emoji-Lora-r4_160282)",
        "",
        "[Link](https://civitai.com/models/177820/sdxl-simple-drawing)",
        "[Link](https://civitai.com/models/221240/texta-generate-text-with-sdxl)",
        "",
    ],
}

df = pd.DataFrame(data)
print(df)
# with open("result.md", "w") as file:
#     file.write(df.to_markdown(index=False))



import time
from pathlib import Path

import pandas as pd
import safetensors.torch
import torch
from diffusers import DiffusionPipeline
from diffusers.utils.constants import USE_PEFT_BACKEND

from onediff.infer_compiler import oneflow_compile
from onediff.torch_utils import TensorInplaceAssign
from onediffx.lora import load_and_fuse_lora, set_and_fuse_adapters, unfuse_lora

if not USE_PEFT_BACKEND:
    raise RuntimeError(
        "The profile if for PEFT APIs, please make sure you have installed peft>=0.6.0 and transformers >= 4.34.0"
    )


class TimerContextManager:
    def __init__(self, msg, lora):
        self.msg = msg
        self.lora_name = lora

    def __enter__(self):
        self.start_time = time.perf_counter()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.end_time = time.perf_counter()
        elapsed_time = self.end_time - self.start_time
        global _time
        _time = elapsed_time
        print(
            f"Time cost {elapsed_time:.2f}, of method {self.msg}, lora name {self.lora_name}"
        )


MODEL_ID = "/share_nfs/hf_models/stable-diffusion-xl-base-1.0"
pipe = DiffusionPipeline.from_pretrained(
    MODEL_ID, variant="fp16", torch_dtype=torch.float16
).to("cuda")

LORA_MODEL_ID = [
    "/share_nfs/onediff_ci/diffusers/loras/SDXL-Emoji-Lora-r4.safetensors",
    "/share_nfs/onediff_ci/diffusers/loras/sdxl_metal_lora.safetensors",
    "/share_nfs/onediff_ci/diffusers/loras/simple_drawing_xl_b1-000012.safetensors",
    "/share_nfs/onediff_ci/diffusers/loras/texta.safetensors",
    "/share_nfs/onediff_ci/diffusers/loras/watercolor_v1_sdxl_lora.safetensors",
]

loras = {name: safetensors.torch.load_file(name) for name in LORA_MODEL_ID}
pipe.unet = oneflow_compile(pipe.unet)
generator = torch.manual_seed(0)

# for OneDiffX APIs
for i, (name, lora) in enumerate(loras.items()):
    load_and_fuse_lora(
        pipe,
        lora.copy(),
        adapter_name=Path(name).stem,
        lora_scale=1.0,
        offload_device="cuda",
    )
    unfuse_lora(pipe)

multi_loras = []
for lora in loras:
    lora = Path(lora).stem
    if len(multi_loras) == 0:
        multi_loras.append([lora])
    else:
        multi_loras.append(multi_loras[-1] + [lora])

set_adapter_time = []
for i, multi_lora in enumerate(multi_loras):
    with TimerContextManager("set_adapter", multi_lora):
        set_and_fuse_adapters(pipe, multi_lora, [0.5] * len(multi_lora))
        unfuse_lora(pipe)
    set_adapter_time.append(_time)

# for PEFT APIs
pipe = DiffusionPipeline.from_pretrained(
    MODEL_ID, variant="fp16", torch_dtype=torch.float16
).to("cuda")
for i, (name, lora) in enumerate(loras.items()):
    pipe.load_lora_weights(lora.copy(), adapter_name=Path(name).stem, lora_scale=1.0)

peft_set_adapter_time = []
for i, multi_lora in enumerate(multi_loras):
    with TimerContextManager("peft set_adapter", multi_lora):
        pipe.set_adapters(multi_lora, [0.5] * len(multi_lora))
        pipe.fuse_lora(adapter_names=multi_lora)
        pipe.unfuse_lora()
    peft_set_adapter_time.append(_time)

lora_names = [
    [1],
    [1, 2],
    [1, 2, 3],
    [1, 2, 3, 4],
    [1, 2, 3, 4, 5],
]

data = {
    "LoRA names": lora_names,
    "PEFT set_adapter": [f"{x:.2f} s" for x in peft_set_adapter_time],
    "OneDiffX set_adapter": [f"{x:.2f} s" for x in set_adapter_time],
}
df = pd.DataFrame(data)
print(df)

with open("result.md", "w") as file:
    file.write(df.to_markdown(index=False))



from compile import OneDiffCompiledGraph

current_unet_graph = OneDiffCompiledGraph()
current_quantization = False
previous_unet_type = {
    "is_sdxl": False,
    "is_sd2": False,
    "is_sd1": False,
    "is_ssd": False,
}
onediff_enabled = False
onediff_backend = None

# controlnet
controlnet_enabled = False
controlnet_compiled = False
previous_is_controlnet = False



import os
from dataclasses import dataclass
from functools import partial, singledispatchmethod
from typing import Any, Dict, Optional, Union

import torch
import torch.nn as nn
from comfy.controlnet import ControlNet
from comfy.model_patcher import ModelPatcher
from onediff.infer_compiler import oneflow_compile
from onediff.infer_compiler.backends.oneflow import (
    OneflowDeployableModule as DeployableModule,
)
from onediff.optimization import quant_optimizer
from onediff_quant.quantization import QuantizationConfig
from onediff_quant.quantization.module_operations import get_sub_module
from onediff_quant.quantization.quantize_calibrators import (
    QuantizationMetricsCalculator,
)
from onediff_quant.quantization.quantize_config import Metric

from .booster_basic import BoosterExecutor
from .patch_management import create_patch_executor, PatchType
from .utils.booster_utils import (
    is_fp16_model,
    set_compiled_options,
    set_environment_for_svd_img2vid,
)
from .utils.graph_path import generate_graph_path


class SubQuantizationPercentileCalculator(QuantizationMetricsCalculator):
    def __init__(
        self,
        model: torch.nn.Module,
        config: QuantizationConfig,
        cache_key: str = "",
        module_selector: callable = lambda x: x,
        *,
        seed=1,
        select=Metric.MAE.value,
        conv_percentage=0.9,
        linear_percentage=0.9,
    ):
        super().__init__(model, config, cache_key, module_selector, seed=seed)
        self.select = select
        self.conv_percentage = conv_percentage
        self.linear_percentage = linear_percentage

    @torch.no_grad()
    def calibrate(self, *args: Any, **kwargs: Any) -> Dict[str, Dict[str, float]]:
        if self.conv_percentage == 0.0 and self.linear_percentage == 0.0:
            return {}
        if (
            (self.conv_percentage == 1.0 and self.linear_percentage == 1.0)
            or (self.conv_percentage == 0.0 and self.linear_percentage == 1.0)
            or (self.conv_percentage == 1.0 and self.linear_percentage == 0.0)
        ):
            # only_use_compute_density
            costs_calibrate_info = self.compute_quantization_costs(
                args, kwargs, module_selector=self.module_selector
            )
            costs_calibrate_info = self.apply_filter(costs_calibrate_info)
            if self.config.cache_dir is not None:
                self.save_quantization_status(
                    costs_calibrate_info, "quantization_stats.json"
                )
            return costs_calibrate_info

        calibrate_info = self.calibrate_all_layers(
            args, kwargs, module_selector=self.module_selector
        )

        selected_model = self.module_selector(self.model)

        # Initialize max and min values, as well as lists for linear and convolutional layer data
        max_value, min_value = 2, -1
        linear_values, conv_values = [max_value, min_value], [max_value, min_value]
        # Iterate through quantization information for each layer, extracting quantization values for linear and convolutional layers
        for module_name, value_info in calibrate_info.items():
            module = get_sub_module(selected_model, module_name)
            values_list = (
                linear_values if isinstance(module, nn.Linear) else conv_values
            )
            values_list.append(float(value_info[self.select]))

        # Sort quantization values for linear and convolutional layers based on the selected evaluation metric
        linear_values.sort()
        conv_values.sort()

        # Calculate linear and convolutional thresholds
        conv_threshold = conv_values[int((len(conv_values) - 1) * self.conv_percentage)]
        linear_threshold = linear_values[
            int((len(linear_values) - 1) * self.linear_percentage)
        ]
        # print(f"Conv threshold: {conv_threshold}, Linear threshold: {linear_threshold}")

        if nn.Conv2d in self.config.module_settings:
            self.config.module_settings[nn.Conv2d][self.select] = conv_threshold

        if nn.Linear in self.config.module_settings:
            self.config.module_settings[nn.Linear][self.select] = linear_threshold

        # Apply filters and save quantization status information
        calibrate_info = self.apply_filter(calibrate_info)
        if self.config.cache_dir is not None:
            self.save_quantization_status(calibrate_info, "quantization_stats.json")

        return calibrate_info


@dataclass
class OnelineQuantizationBoosterExecutor(BoosterExecutor):
    conv_percentage: int = 60
    linear_percentage: int = 70
    conv_compute_density_threshold: int = 100
    linear_compute_density_threshold: int = 300

    @singledispatchmethod
    def execute(self, model, ckpt_name=None, **kwargs):
        print(f"{type(self).__name__}.apply() not implemented for {type(model)}")
        return model

    @singledispatchmethod
    def extract_torch_module(self, model):
        raise NotImplementedError(f"{type(model)}")

    @execute.register(ModelPatcher)
    @execute.register(ControlNet)
    def _(
        self, model: ModelPatcher, ckpt_name: Optional[str] = None, **kwargs
    ) -> Union[ModelPatcher, ControlNet]:
        # Define quantization settings
        quant_config = QuantizationConfig.from_settings(
            quantize_conv=True,
            quantize_linear=True,
            bits=8,
            conv_mae_threshold=0.9,
            linear_mae_threshold=0.9,
            plot_calibrate_info=True,
            conv_compute_density_threshold=self.conv_compute_density_threshold,
            linear_compute_density_threshold=self.linear_compute_density_threshold,
        )

        def get_torch_module(module):
            if isinstance(module, DeployableModule):
                return module._torch_module
            return module

        # Retrieve the model to be optimized
        def get_target_model(model, callback=None):
            if isinstance(model, ModelPatcher):
                target_model = model.model.diffusion_model
            elif isinstance(model, ControlNet):
                target_model = model.control_model
            else:
                raise NotImplementedError

            if callback:
                return callback(target_model)

            return target_model

        def set_optimized_model(model, quant_config):
            target_model = get_target_model(model)
            torch_model = get_torch_module(target_model)
            if isinstance(target_model, DeployableModule):
                compiled_model = target_model
            else:
                compiled_model = oneflow_compile(target_model)

            if isinstance(model, ModelPatcher):
                model.model.diffusion_model = compiled_model
            elif isinstance(model, ControlNet):
                model.control_model = compiled_model
            else:
                raise NotImplementedError

            compiled_model.apply_online_quant(quant_config)
            if ckpt_name:
                cache_key = f"{ckpt_name}_{type(torch_model).__name__}"
                graph_file = generate_graph_path(cache_key, torch_model)
                quant_config.cache_dir = os.path.dirname(graph_file)
                set_compiled_options(compiled_model, graph_file)

        set_environment_for_svd_img2vid(model)
        # Get the Torch model and set quantization calculator
        torch_model = get_target_model(model, callback=get_torch_module)
        if not is_fp16_model(torch_model):
            print(
                f"Warning: Model {type(torch_model)} is not an FP16 model. Compilation will be skipped!"
            )
            return model

        quant_config.quantization_calculator = SubQuantizationPercentileCalculator(
            torch_model,
            quant_config,
            cache_key=type(torch_model).__name__,
            conv_percentage=self.conv_percentage / 100,
            linear_percentage=self.linear_percentage / 100,
        )

        # Apply quantization optimization and return the optimized model
        if create_patch_executor(PatchType.DCUNetExecutorPatch).check_patch(model):
            self._set_optimized_model_for_deepcace(model)
        else:
            set_optimized_model(model, quant_config)
        return model

    def _set_optimized_model_for_deepcace(self, model: ModelPatcher):
        # TODO
        print(
            "Warning: DeepCache + OnelineQuantization only support default configurations:"
        )
        model.fast_deep_cache_unet.quantize = partial(
            quant_optimizer.quantize_model,
            model.fast_deep_cache_unet,
            quantize_conv=False,
        )
        model.deep_cache_unet.quantize = partial(
            quant_optimizer.quantize_model, model.deep_cache_unet, quantize_conv=False
        )



from functools import partial, singledispatchmethod
from typing import Optional

import torch

from comfy.controlnet import ControlLora, ControlNet
from comfy.model_patcher import ModelPatcher
from comfy.sd import VAE
from nexfort.utils.memory_format import apply_memory_format

from onediff.infer_compiler import compile

from ..booster_interface import BoosterExecutor
from .onediff_controlnet import OneDiffControlLora


class BasicNexFortBoosterExecutor(BoosterExecutor):
    # https://pytorch.org/docs/stable/_modules/torch.html#compile
    def __init__(
        self,
        mode: str = "max-optimize:max-autotune:low-precision",
        fullgraph=False,
        dynamic=True,
    ):
        super().__init__()
        options = {
            "mode": mode,
            "dynamic": dynamic,
            "fullgraph": fullgraph,
        }
        self.compile_fn = partial(compile, backend="nexfort", options=options)
        self.options = options

    @singledispatchmethod
    def execute(self, model, ckpt_name=None, **kwargs):
        raise NotImplementedError(f"Cannot execute {type(model)=}")

    @execute.register(ModelPatcher)
    @torch.inference_mode()
    def _(self, model, ckpt_name: Optional[str] = None, **kwargs):
        diffusion_model = model.model.diffusion_model
        model.model.diffusion_model = apply_memory_format(
            diffusion_model, torch.channels_last
        )
        model.model.diffusion_model = self.compile_fn(diffusion_model)
        model.weight_inplace_update = True
        return model

    @execute.register(VAE)
    @torch.inference_mode()
    def _(self, model, ckpt_name: Optional[str] = None, **kwargs):
        model.first_stage_model = apply_memory_format(
            model.first_stage_model, torch.channels_last
        )
        print(f"{type(model)} apply compiled config: {self.options}")
        # https://huggingface.co/blog/sd3#performance-optimizations-for-sd3
        model.first_stage_model.decode = self.compile_fn(model.first_stage_model.decode)
        return model

    @execute.register(ControlNet)
    @torch.inference_mode()
    def _(self, model, ckpt_name: Optional[str] = None, **kwargs):
        torch_model = model.control_model
        torch_model = apply_memory_format(torch_model, torch.channels_last)
        compiled_model: torch.nn.Module = self.compile_fn(torch_model)
        model.control_model = compiled_model
        return model

    @execute.register(ControlLora)
    @torch.inference_mode()
    def _(self, model, ckpt_name: Optional[str] = None, **kwargs):
        def compile_cnet(model):
            out: torch.nn.Module = self.compile_fn(model)
            return out

        model = OneDiffControlLora.from_controllora(model, compile_fn=compile_cnet)
        return model
